import jwt

from nsanic.libs.tool_dt import cur_time


def jencode(
        user_id: int or str,
        jw_type: int,
        safe_key='qazxswedc1234',
        client_info='',
        extra=None,
        useful_life: int = None
):
    """
    JWT加密 默认采用HS256

    :param user_id: 用户ID
    :param jw_type: 授权类型 暂时分四大块： 管理账户、代销账户、常规账户、线下设备
    :param safe_key: 用户随机授权码，一般在用户刷新jwt token、注销登录等操作后会绑定唯一设备等操作时会重置，保证token唯一性
    :param client_info: 客户端唯一标识信息(可选),设置该值可以标识token的唯一性
    :param extra: (可选)授权附加信息,注入后解析可直接获得 勿放置敏感信息
    :param useful_life: token 有效期
    """
    cur_t = cur_time()
    ex_time = (useful_life or 7 * 86400) + cur_t
    preload = {
        'sub': jw_type,
        'iss': 'default',
        'exp': ex_time,
        'iat': cur_t,
        'aud': f'{jw_type}_{client_info}' if client_info else str(jw_type),
        'identify': user_id,
        'extra': extra,
    }
    head_dict = {
        'typ': 'JWT',
        'alg': 'HS256',
    }
    jwt_str = jwt.encode(preload, key=safe_key, headers=head_dict)
    return jwt_str


def get_jwinfo(jwt_str: str):
    """
    获取jwt token 加配信息 账户ID, 授权类型, 附加信息
    """
    try:
        data = jwt.decode(jwt_str, algorithms=['HS256'], options={'verify_exp': False, 'verify_signature': False})
    except (jwt.exceptions.DecodeError, jwt.exceptions.InvalidTokenError):
        return None, 'Invalid authorization'
    if not data:
        return None, 'Invalid authorization'
    uid, jw_type = data.get('identify'), data.get('sub')
    if (not uid) or (not jw_type):
        return None, 'Invalid authorization'
    return uid, data


def jdecode(jwt_str: str, jw_type: int, safe_key='qazxswedc1234', client_info='', new_check=False):
    """
    JWT TOKEN 解析
    """
    try:
        client_info = f'{jw_type}_{client_info}' if client_info else str(jw_type)
        info = jwt.decode(jwt_str, key=safe_key, algorithms=['HS256'], audience=client_info)
    except (jwt.exceptions.DecodeError, jwt.exceptions.InvalidTokenError, jwt.exceptions.InvalidKeyError):
        return None, 'Invalid authorization'
    except jwt.exceptions.InvalidSignatureError:
        return None, 'Error certificate'
    except jwt.exceptions.InvalidAudienceError:
        return None, 'Authorization not match'
    except jwt.exceptions.ExpiredSignatureError:
        return None, 'Authorization expired'
    cur_t, new_token = cur_time(), None
    if new_check and (info.get('exp') - cur_t <= 43200):
        # token有效期低于1天的自动刷新token
        new_token = jencode(
            user_id=info.get('identify'), jw_type=jw_type, safe_key=safe_key,
            client_info=client_info, extra=info.get('extra'))
    return info, new_token



import os
import random
import re
import string


def input_fun(check_fun, tips='请输入：'):
    while 1:
        item_val = input(tips)
        val = check_fun(item_val)
        if val:
            return val


def check_name(val: str):
    val = val.strip()
    if not val:
        print('必须指定服务名')
        return
    val = val.lower()
    if not re.match(r'^[a-z][a-z0-9_]*[a-z0-9]$', val):
        print('服务命名不符合规范，命名只能包含数字/字母/下划线，且不能以数字或下划线开头或结尾')
        return
    if val.find('__') != -1:
        print('服务命名不符合规范，命名不能包含空格或多个下划线')
        return
    return val


def check_id(val: str):
    val = val.strip()
    val = val.upper() if val else f"{''.join(random.sample(string.ascii_uppercase, 1))}{''.join(random.sample(string.digits, 4))}"
    if 3 <= len(val) <= 6 and re.match(r'^[A-Z][A-Z0-9]+[0-9]$', val):
        return val
    print('服务ID应是三到六位字母开头数字结尾组成的字符串，且应具有唯一性')


def check_db(val: str):
    val = val.strip()
    val = val.lower().strip() if val else 'y'
    if val in ('y', 'n'):
        return val
    print('是否使用数据库输入参数不符合规范，请输入y或n')


def check_m_list(val: str):
    val = val.strip()
    if not val:
        return ['main']
    val_list = val.split(';')
    new_list = []
    for v in val_list:
        v = check_name(v)
        if v:
            new_list.append(v)
        else:
            return
    return new_list


def check_redis(val: str):
    val = val.strip()
    val = val.lower() if val else 'y'
    if val in ('y', 'n'):
        return val
    print('是否使用redis缓存输入参数不符合规范，请输入y或n')


def check_port(val: str):
    val = val.strip()
    if not val:
        return 8888
    if val.isdigit() and (80 <= int(val) <= 65535):
        return int(val)
    print('输入的端口号应是数字类型且在80到65535之间')


def check_use_fast(val):
    val = val.strip()
    val = val.lower() if val else 'y'
    if val in ('y', 'n'):
        return val
    print('是否使用自动服务进程配置参数不符合规范，请输入y或n')


def check_ctrl_version(val):
    val = val.strip()
    val = val.lower() if val else 'n'
    if val == 'n':
        return 'n'
    if len(val) > 10:
        print('接口默认版本参数不能超过10个字符长度')
        return None
    if re.match(r'^[a-zA-Z0-9][a-zA-Z0-9.]*[a-zA-Z0-9]$', val):
        return val
    print('是否使用接口版本控制参数只能在数字、大小写字母和点之间选择，并且已数字或字母开头、以及结尾')


engine_map = {
    '1': ('tortoise.backends.mysql', 'python -m pip install tortoise-orm[aiomysql]'),
    '2': ('tortoise.backends.asyncpg', 'python -m pip install tortoise-orm[asyncpg]'),
    # '3': ('tortoise.backends.aiosqlite', 'python -m pip install tortoise-orm[aiosqlite]'),
    # '4': ('tortoise.backends.asyncodbc', 'python -m pip install tortoise-orm[asyncodbc]'),
}


def check_use_engine(val):
    val = val.strip()
    if not val:
        return '2'
    if val in engine_map:
        return val
    print('无效的数据引擎指定')


service_map = {
    1: 'web',
    2: 'ws',
    3: 'srv',
}


def check_use_mode(val):
    val = val.strip()
    if not val:
        return 1
    if val in ('1', '2', '3'):
        return int(val)
    print('无效的服务模式指定')


def check_use_sub(val):
    val = val.strip()
    val = val.lower() if val else 'y'
    if val in ('y', 'n'):
        return val
    print('无效的websocket消息订阅模式')


def add_file(file_path: str, content=''):
    with open(file_path, 'w', encoding='utf-8') as f:
        if content:
            f.write(content)


def add_module(path: str, name: str, init_into=''):
    m_path = os.path.join(path, name)
    try:
        os.makedirs(m_path)
        add_file(os.path.join(m_path, '__init__.py'), content=init_into)
    except FileExistsError:
        print(f'指定的文件或目录已存在，不会重复创建：{m_path}')
        return
    return m_path


s_name = input_fun(check_name, tips='服务名称(建议小写并用下划线连接，长度不超过20字)：')
base_path = os.getcwd()
server_path = add_module(base_path, s_name)
if not server_path:
    raise FileExistsError('指定服务已存在，不能重复创建')
s_id = input_fun(check_id, tips='服务ID(三到六位，不区分大小写，默认随机)：')
use_db = input_fun(check_db, tips='是否使用数据库(y(默认)/n)：')
use_engine = input_fun(check_use_engine, tips='选择数据库存储引擎(1-mysql/2-pgsql(默认))：')
m_list = input_fun(check_m_list, tips='指定迁移模型文件名(多个用分号隔开，默认main)：')
o_list = input_fun(check_m_list, tips='指定不做迁移的模型文件名(多个用分号隔开，默认extra)：')
use_redis = input_fun(check_redis, tips='是否使用redis缓存(y(默认)/n,启用包含websocket的服务则该项自动设置为y)：')
port = input_fun(check_port, tips='服务运行端口号(80到65535之间的数字，默认8888)：')
use_fast = input_fun(check_use_fast, tips='是否自动配置服务进程(y(默认)/n)：')
use_ver = input_fun(check_ctrl_version, tips='接口版本版本控制(n(默认不启用)/数字字母或点的任意组合)：')
use_mode = input_fun(check_use_mode, tips='指定服务模式(1-http(默认),2-websocket,3-混合模式)：')
use_sub = 'y'
if use_mode in (2, 3):
    use_sub = input_fun(check_use_sub, tips='是否启用Websocket消息订阅模式(y(默认)/n)：')

dir_config = add_module(server_path, 'config')
dir_db = add_module(server_path, 'model_db')
dir_cache = add_module(server_path, 'model_rc')
dir_api = add_module(server_path, 'interface')
dir_handler = add_module(server_path, 'handler')
dir_script = add_module(server_path, 'script')

class_conf = f"ConfSrv"
db_val = None if use_db != 'y' else '''{
        'default': {
            'engine': '%s', 
            'credentials': {
                'host': '127.0.0.1', 
                'port': '5432', 
                'user': '', 
                'password': '', 
                'database': ''
            },
            'options': {
                'charset': 'utf8mb4', 
                'minsize': 5, 
                'maxsize': 30
            }
        }
    }''' % engine_map.get(use_engine)[0]
rds_val = None if (use_mode == 1) and use_redis != 'y' else '''{
        'default': {
            'host': '127.0.0.1', 
            'port': 6379, 
            'password': '', 
            'db': 1, 
            # 'max_connections': 5,  # 最大连接数 进程开多按情况降低连接数 最小值为2
            # 'decode_responses': True,  # 默认解析字符串，不解析默认为bytes类型
            'socket_connect_timeout': 5,  # 连接超时时长
            # ‘timeout’: 10,  # 重启连接间隔以及等待连接释放的超时时长
            # ‘ping_interval’: 30,  # ping间隔
            # 'use_block': False,   # 是否使用阻塞式共享连接池管理，阻塞式连接池管理能相对精准管控最大连接数，节省连接资源，但性能有所下降
        }, 
    }'''
conf_data = f'''# coding=utf-8
from nsanic.base_conf import BaseConf


class {class_conf}(BaseConf):
    SERVER_NAME = '{s_name}'
    SERVER_ID = '{s_id}'
    RUN_PORT = {port}
    DEBUG_MODE = False
    ACCESS_LOG = False
    MODEL_LIST = {m_list}
    RUN_FAST = {'False' if use_fast != 'y' else 'True'}
    VER_CODE = {None if use_ver == 'n' else use_ver}
    MODEL_EXTRA = {o_list}
    CONF_DB = {db_val}
    CONF_RDS = {rds_val}
'''
if m_list == ['main']:
    conf_data = conf_data.replace("MODEL_LIST = ['main']", '')
    add_file(os.path.join(dir_db, 'main.py'))
else:
    for dbf in m_list:
        add_file(os.path.join(dir_db, f'{dbf}.py'))
if o_list == ['main']:
    conf_data = conf_data.replace("MODEL_EXTRA = ['main']", '')
    add_file(os.path.join(dir_db, 'extra.py'))
else:
    for dbf in o_list:
        add_file(os.path.join(dir_db, f'{dbf}.py'))
add_file(os.path.join(dir_config, 'conf_start.py'), content=conf_data)
conf_key = s_name.split('_')[0]
init_data = f'''# coding=utf-8
from {s_name}.config.conf_start import {class_conf}


conf_srv = {class_conf}()
migrate_db = conf_srv.migrate_db()
'''
add_file(os.path.join(dir_config, '__init__.py'), content=init_data)

enum_data = '''# coding=utf-8
from nsanic.libs.consts import BaseEnum


# 请在此处创建枚举对象

'''
add_file(os.path.join(dir_config, 'consts.py'), content=enum_data)

ws_class = 'WsRdsConnector' if use_sub == 'y' else 'WsConnector'
connector_str = """WsProtocol.connector_map = {{
    {}.__name__: {}
}}
""".format(ws_class, ws_class)
ws_model = f'''from nsanic.base_ws import BaseWebsocket

from nsanic.libs.manager import {ws_class}''' if use_mode in (2, 3) else ''
ws_base = f'''\n\n\nclass BaseWebsocketApi(BaseWebsocket):

    conf = conf_srv
    ws_manager = {ws_class}''' if use_mode in (2, 3) else ''
api_data = f'''# coding=utf-8
from nsanic.handler_http import BaseHttpApi
{ws_model}
from {s_name}.config import conf_srv, ConfSrv


class BaseApi(BaseHttpApi):
    pass
'''
add_file(os.path.join(server_path, 'base_api.py'), content=api_data)

test_data = f'''# coding=utf-8
from sanic import Request
from {s_name}.base_api import BaseApi


class TestApi(BaseApi):
    async def get(self, req: Request):
        return self.answer(hint="Here is a api for test.")
'''
add_file(os.path.join(dir_api, 'test_api.py'), content=test_data)

url_data = f'''# coding=utf-8
from nsanic.base_blue import BaseBlue
from nsanic.handler_http import Urls
from {s_name}.interface.test_api import TestApi


class MainBp(BaseBlue):
    # 路由请添加在这里
    DEFAULT_APIS = [
        # 结构为: 接口路由地址, 接口视图处理器, 版本号(可选), 接口命名(可选)
        Urls("/testapi", TestApi),
    ]
'''
add_file(os.path.join(server_path, 'url_main.py'), content=url_data)
ws_proto = 'from nsanic.base_ws import WsProtocol' if use_mode in (2, 3) else ''
ws_run = 'main_server.run(protocol=WsProtocol)' if use_mode in (2, 3) else 'main_server.run()'
connector = "from nsanic.libs.manager import {}".format(ws_class) if use_mode in (2, 3) else ""
start_data = '''# coding=utf-8
from nsanic.base_server import InitServer
{4}
from nsanic.exception import CatchExpt
from nsanic.middleware import CorsMiddle
{1}
from {0}.config import conf_srv as conf
from {0}.url_main import MainBp

signal_map = {{}}
{3}
main_server = InitServer(conf, mws=[CorsMiddle], bps=[MainBp], excps=[CatchExpt])
main_server.add_signal(signal_map)


if __name__ == '__main__':
    {2}
'''.format(s_name, ws_proto, ws_run, connector_str if use_mode in (2, 3) else '', connector)
add_file(os.path.join(base_path, f'{service_map.get(use_mode)}_{s_name}.py'), content=start_data)
if use_db == 'y':
    os.system(engine_map.get(use_engine)[1])
    os.system('pip install aerich')



from dataclasses import dataclass
from typing import Annotated, Union

from sanic import Request
from sanic.views import HTTPMethodView

from nsanic import verify
from nsanic.base_ws import BaseWebsocket
from nsanic.exception import JsonFinish
from nsanic.libs.component import LogMeta
from nsanic.libs.consts import Code, StaCode
from nsanic.libs.tool_dt import to_datetime


@dataclass
class Urls:
    """接口路由对象"""
    router: str
    '''路由'''
    handler: Annotated[Union[HTTPMethodView, BaseWebsocket, type(HTTPMethodView), type(BaseWebsocket)], 'HTTPMethodView or subclass', 'BaseWebsocket or subclass']
    '''API视图处理器'''
    ver: str = ''
    '''接口版本号'''
    name: str = ''
    '''接口名称'''


class BaseRps(LogMeta):
    STA_CODE = StaCode

    @classmethod
    def check_int(
            cls,
            val,
            require=False,
            default: int = None,
            minval: int = None,
            maxval: int = None,
            inner=True,
            p_name='') -> int or None:
        """
        整形参数校验
        :param val: 待校验值
        :param require: 是否必要参数 默认非必要
        :param default: 非必要状态下的默认值
        :param minval: 最小值范围 type--int 默认不校验
        :param maxval: 最大值范围 type--int 默认不校验
        :param inner: 是否范围内(针对于min_val或max_val有值校验)，默认范围内，False为范围外
        :param p_name: 参数名
        :return 转换的值--int
        """
        sta, val_info = verify.vint(
            val, require=require, default=default, minval=minval, maxval=maxval, inner=inner, p_name=p_name)
        return val_info if sta else cls.answer(code=StaCode.ERR_ARG, hint=val_info)

    @classmethod
    def check_str(
            cls,
            val,
            require=False,
            default: str = None,
            turn=0,
            minlen: int = None,
            maxlen: int = None,
            p_name='') -> str or None:
        """
        字符串参数校验
        :param val: 待校验对象
        :param require: 是否必要参数 默认非必要
        :param default: 非必要状态下的默认值
        :param turn: 默认转换 1--转化大写 2--转换小写 其它值--不转化
        :param minlen: 最小长度 type--int 默认不校验
        :param maxlen: 最大长度 type--int 默认不校验
        :param p_name: 参数名
        :return 转换的值--str
        """
        sta, val_info = verify.vstr(
            val, require=require, default=default, turn=turn, minlen=minlen, maxlen=maxlen, p_name=p_name)
        return val_info if sta else cls.answer(code=StaCode.ERR_ARG, hint=val_info)

    @classmethod
    def check_float(
            cls,
            val,
            require=False,
            default: (float, int) = None,
            keep_val=3,
            minval: float or int = None,
            maxval: float or int = None,
            inner=True,
            p_name='') -> float or None:
        """
        浮点数校验
        :param val: 待校验对象
        :param require: 是否必要参数 默认非必要
        :param default: 非必要状态下的默认值
        :param keep_val: 保留小数未 默认3位
        :param minval: 最小值范围 type--int 默认不校验
        :param maxval: 最大值范围 type--int 默认不校验
        :param inner: 是否范围内(针对于min_val或max_val有值校验)，默认范围内，False为范围外
        :param p_name: 参数名
        :return 转换的值--float
        """
        sta, val_info = verify.vfloat(
            val, require=require, default=default, keep_val=keep_val, minval=minval, maxval=maxval,
            inner=inner, p_name=p_name)
        return val_info if sta else cls.answer(code=cls.STA_CODE.ERR_ARG, hint=val_info)

    @classmethod
    def check_time(cls, val, require=False, default=None, time_min=None, time_max=None, inner=True, p_name=''):
        sta, val_info = verify.vdatetime(
            val, require=require, default=default, time_min=time_min, time_max=time_max, inner=inner, p_name=p_name)
        return val_info if sta else cls.answer(code=cls.STA_CODE.ERR_ARG, hint=val_info)

    @classmethod
    def check_type(cls, val, query_fun, require=False, default=None, is_int=True, p_name=''):
        val = cls.check_int(val, require=require, default=default, p_name=p_name) if is_int else \
            cls.check_str(val, require=require, default=default, p_name=p_name)
        if val is None:
            return
        if query_fun(val):
            return val
        return cls.answer(code=cls.STA_CODE.ERR_ARG, hint=f'{p_name}参数不在限定范围')

    @classmethod
    def check_phone_number(cls, phone_str: str, require=False):
        phone_str = cls.check_str(phone_str, require=require, p_name='手机号')
        if phone_str:
            if verify.phone(phone_str):
                return phone_str
            return cls.answer(code=cls.STA_CODE.ERR_ARG, hint='手机号不符合规范')

    @classmethod
    def check_page(cls, req: Request, page_max=200):
        """
        公共校验分页参数

        :param req: sanic请求对象
        :param page_max: 分页最大限制
        :return 校验状态, 分页偏移量, 分页限制量
        """
        page, page_size = (req.args.get('page'), req.args.get('page_size')) if req.method == 'GET' else (
            req.json.get('page'), req.json.get('page_size'))
        page = cls.check_int(page, default=1, minval=1, p_name='页码')
        page_size = cls.check_int(page_size, default=20, minval=1, maxval=page_max, p_name='分页大小')
        return page, page_size

    @classmethod
    def check_time_inner(cls, req: Request, begin_str='beginTime', end_str='endTime'):
        """检查时间区间 并按指定的查询参数集补充时间区间查询"""
        s_time, e_time = (req.args.get(begin_str), req.args.get(end_str)) if req.method == 'GET' else (
            req.json.get(begin_str), req.json.get(end_str))
        if s_time:
            s_time_str = cls.check_str(s_time, require=True, minlen=10, maxlen=20, p_name=begin_str)
            if len(s_time_str) <= 10:
                s_time_str += " 00:00:00"
            s_time = to_datetime(s_time_str)
        if e_time:
            e_time_str = cls.check_str(e_time, require=True, minlen=10, maxlen=20, p_name=end_str)
            if len(e_time_str) <= 10:
                e_time_str += " 00:00:00"
            e_time = to_datetime(e_time_str)
        s_time and e_time and e_time <= s_time and cls.answer(code=cls.STA_CODE.ERR_ARG,
                                                              hint='开始时间至少要大于结束时间')
        return s_time or None, e_time or None

    @classmethod
    def answer(
            cls,
            code: Code = None,
            data: (dict, object, list) = None,
            total: int = 0,
            hint: str = '',
            headers: dict = None):
        """
        公共JSON响应函数PG游戏联调

        :param code: 响应码,请参照StaCode中取值, 默认响应成功状态
        :param data: 响应数据, 可以是任意符合JSON规范类型的数据模型
        :param total: 针对于分页响应的总数量
        :param hint: 响应消息, 字符串, 设置值后会采取设置的值，否则会使用响应码映射的默认值
        :param headers: 附加响应头
        """
        if not code:
            code = cls.STA_CODE.PASS
        raise JsonFinish(code, data, total, hint, headers)


class BaseHttpApi(BaseRps, HTTPMethodView):
    """
    公用请求基础类
    """



import asyncio
from datetime import datetime, date
from enum import Enum
from typing import Union
from uuid import UUID

from tortoise import Model, fields, connections, transactions
from tortoise.converters import encoders
from tortoise.expressions import Q

from nsanic.libs import tool_dt

GeneratorClass = None


class DBModel(Model):
    _SPLIT_TABLES = set()
    '''当前分表集合'''
    _SPLIT_SUFFIX = {1: 'str', 2: 'y%Y', 3: 'y%Ym%m', 4: 'y%Yw%w', 5: 'y%Ym%md%d'}
    '''表名后缀类型映射 1--自定义标记 2--按日期年标记 3--按日期年月标记 4--按日期年周标记 5--按日期年月日标记'''
    _SPLIT_TYPE = 0
    '''分表模式 0--不分表 1--自定义标记分表 2--按年分表 3--按月分表 4--按周分表 5--按日分表'''
    _SPLIT_FIELD = 'created'
    '''分表字段或标记字符串'''

    created = fields.BigIntField(null=True, index=True, default=0, description='创建时间')

    class Meta:
        abstract = True

    @classmethod
    def sheet_name(cls, suffix: Union[str, int, float, datetime, date] = None, tz: str = None):
        """
        数据表名 
        :param suffix: 当前待设置的后缀或后缀数据
        :param tz: 指定按日期分表数据的时区
        """
        tb_name = cls._meta.db_table or cls.__name__.lower()
        if cls._SPLIT_TYPE:
            if cls._SPLIT_TYPE not in cls._SPLIT_SUFFIX:
                raise Exception('Invalid split table type.')
            if (cls._SPLIT_TYPE == 1) or (isinstance(suffix, str)):
                tb_name = f"{tb_name}_{suffix}"
            else:
                fmt = cls._SPLIT_SUFFIX.get(cls._SPLIT_TYPE)
                tb_name = f"{tb_name}_{tool_dt.dt_str(suffix, fmt=fmt, tz=tz)}"
        return tb_name

    @classmethod
    def pk_name(cls):
        return cls._meta.db_pk_column

    @classmethod
    def check_field(cls, field: str):
        return field in cls._meta.db_fields

    @classmethod
    def out_fields(cls, forbids: Union[list, tuple, set] = None):
        """禁止部分字段输出"""
        return [name for name in cls._meta.db_fields if name not in forbids] if forbids else cls._meta.db_fields

    @staticmethod
    def check_val(val, field, model):
        if val is not None:
            if isinstance(val, UUID):
                val = str(val)
            elif isinstance(val, Enum):
                val = val.val if hasattr(val, 'val') else val.value
            return encoders.get(type(val))(val)
        column = model._meta.fields_map.get(field)
        if not column:
            raise Exception(f'数据模型{model.__name__}不包含指定的字段{field}')
        vdft = column.default
        if vdft is not None:
            if isinstance(vdft, UUID):
                vdft = str(vdft)
            elif isinstance(vdft, Enum):
                vdft = vdft.val if hasattr(vdft, 'val') else vdft.value
            return encoders.get(type(vdft))(vdft)
        return "NULL"

    @classmethod
    def get_meta_db(cls, db_key: str = None):
        if (not db_key) or (db_key == 'default'):
            return cls._meta.db
        else:
            conn = connections.get(db_key)
            if not conn:
                raise Exception(f'Can not found {db_key} db config, please confirm the config is exist.')
        if type(conn).__name__ == 'TransactionWrapper':
            return conn._parent
        return conn

    @classmethod
    def __check_orders(cls, orders: Union[str, list, tuple]):
        if orders:
            if isinstance(orders, str):
                orders = [orders]
            new_order = []
            for key in orders:
                key_str = key[1:] if key[0] in ('-', '+') else key
                (key_str in cls._meta.db_fields) and new_order.append(key)
            return new_order
        return []

    @classmethod
    def __fetch_generator_class(cls):
        global GeneratorClass
        if GeneratorClass:
            return GeneratorClass
        dialect = cls._meta.db.schema_generator.DIALECT
        if dialect == 'postgres':
            from nsanic.orm.pgsql_generator import PgSqlGenerator
            GeneratorClass = PgSqlGenerator
        elif dialect == 'mysql':
            from nsanic.orm.mysql_generator import MySqlGenerator
            GeneratorClass = MySqlGenerator
        elif dialect == 'mssql':
            from nsanic.orm.mssql_generator import MsSqlGenerator
            GeneratorClass = MsSqlGenerator
        return GeneratorClass

    @classmethod
    def get_sheet_name(cls, tb_suffix: str = None):
        or_name = cls._meta.db_table or cls.__name__.lower()
        return f"{or_name}_{tb_suffix}" if tb_suffix else or_name, or_name

    @classmethod
    def get_split_suffix(cls, split: Union[str, int, float, datetime, date], param: dict, tz: str = None):
        """
        获取字典参数的分表后缀
        :param split: 分表标记或标记位数据
        :param param: 查询参数
        :param tz: 按日期分表数据的时区
        """
        if not cls._SPLIT_TYPE:
            return ''
        if cls._SPLIT_TYPE not in cls._SPLIT_SUFFIX:
            raise Exception('Invalid split table type.')
        if cls._SPLIT_TYPE == 1:
            return f'{split}'
        if not split:
            split = cls._SPLIT_FIELD
        kval = (param.get(split) or param.get(f'{split}__gte') or param.get(f'{split}__lte')
                or param.get(f'{split}__gt') or param.get(f'{split}__lt')) or tool_dt.cur_time()
        return f'{tool_dt.dt_str(kval, fmt=cls._SPLIT_SUFFIX.get(cls._SPLIT_TYPE), tz=tz)}'

    @classmethod
    async def exec_sql(cls, sql: str, query=False, for_one=False, db_key='default'):
        db = cls.get_meta_db(db_key)
        if query:
            sta = await db.execute_query(sql)
        else:
            await db.execute_script(sql)
            sta = [1, [1]]
        if (not sta) or (not sta[0]):
            if query:
                return []
            error = f'执行sql失败--{sta}: {sql}'
            raise Exception(error)
        return (sta[1] if (not for_one) else sta[1][0]) if sta and (len(sta) > 1) else True

    @classmethod
    async def exec_query(cls, sql: str, require=False, retry=1, for_one=False, db_key='default'):
        if (not require) or (retry <= 1):
            return await cls.exec_sql(sql, query=True, for_one=for_one, db_key=db_key)
        for i in range(retry):
            sta = await cls.exec_sql(sql, query=True, for_one=for_one, db_key=db_key)
            if sta:
                return sta
            await asyncio.sleep(0.2)
        return None

    @classmethod
    async def exec_insert(cls, sql_list: list[str], use_transaction=False, db_key='default'):
        db = cls.get_meta_db(db_key)
        if not use_transaction:
            for sql in sql_list:
                sta = await db.execute_query(sql)
                if (not sta) or (not sta[0]):
                    error = f'执行sql失败--{sta}: {sql}'
                    raise Exception(error)
        else:
            async with transactions.in_transaction(db_key or 'default'):
                for sql in sql_list:
                    sta = await db.execute_query(sql)
                    if (not sta) or (not sta[0]):
                        error = f'执行sql失败--{sta}: {sql}'
                        raise Exception(error)
        return True

    @classmethod
    def gen_simple_query(
            cls,
            param: dict = None,
            field: Union[tuple, list, set] = None,
            exclude: Union[tuple, list, set] = None,
            groups: Union[str, list, tuple] = None,
            orders: Union[str, list, tuple] = None,
            offset=0,
            limit=0,
            with_del=False,
            with_field=False,
            db_key: str = None):
        if db_key and (db_key != 'default'):
            db = cls.get_meta_db(db_key)
            q_set = cls.filter(sta_del=False).using_db(db) if (hasattr(cls, 'sta_del')) and (not with_del) else cls.all(using_db=db)
        else:
            q_set = cls.filter(sta_del=False) if (hasattr(cls, 'sta_del')) and (not with_del) else cls.all()
        if param:
            q_set = q_set.filter(Q(**param))
        if groups:
            if isinstance(groups, str):
                groups = [groups]
            q_set = q_set.group_by(*groups)
        orders = cls.__check_orders(orders)
        if orders:
            q_set = q_set.order_by(*orders)
        if limit:
            q_set = q_set.limit(limit)
        if offset:
            q_set = q_set.offset(offset)
        if with_field:
            out_field = [f for f in field if f in cls._meta.db_fields] if field else (
                [v for v in cls._meta.db_fields if v not in exclude] if exclude else [])
            return q_set, out_field
        return q_set

    @classmethod
    async def check_table(cls, suffix: str = None, db_key: str = None):
        tb_name, or_name = cls.get_sheet_name(suffix)
        if tb_name not in cls._SPLIT_TABLES:
            cls._meta.db_table = cls._meta.basetable._table_name = tb_name
            _creator = cls.__fetch_generator_class()
            if not _creator:
                cls._meta.db_table = or_name
                raise Exception(f'未定义的表创建模式：{cls._meta.db.schema_generator.DIALECT}')
            db = cls.get_meta_db(db_key)
            _sql_dict = _creator(db).get_table_sql_new(cls)
            cls._meta.db_table = or_name
            if not _sql_dict:
                raise Exception(f'无法生成建表模型：{_sql_dict}')
            sql = _sql_dict.get('table_creation_string')
            if not sql:
                raise Exception(f'无法生成创建表SQL：{_sql_dict}')
            await cls._meta.db.execute_script(sql)
            cls._SPLIT_TABLES.add(tb_name)

    @classmethod
    async def fetch_on_page(
            cls,
            param: dict = None,
            page=1,
            size=20,
            field: Union[tuple, list, set] = None,
            exclude: Union[tuple, list, set] = None,
            groups: Union[str, list, tuple] = None,
            orders: Union[str, list, tuple] = '-created',
            split: Union[str, int, float, datetime, date] = None,
            with_del=False,
            with_count: int = 1,
            db_key: str = None,
            tz: str = None):
        """
        简单分页查询
        :param param:字典形式的查询参数, __lte __gte 等可自行通过key构造
        :param page: 页数
        :param size: 分页大小
        :param field: 指定输出的字段（优先）
        :param exclude: 指定不输出的字段
        :param groups: 分组字段
        :param orders: 排序字段
        :param split: 分表标记 设置分表类型为1时生效，
        :param with_count: 是否进行数量查询 0-否 1-精确数量 2-模糊数量(针对大表)
        :param with_del: 是否包含逻辑删除数据
        :param db_key: 采用的数据库配置
        :param tz: 指定按日期分表采用的时区
        """
        if page < 1:
            page = 1
        offset = (page - 1) * size
        if cls._SPLIT_TYPE <= 1:
            total = 0
            q_set, out_field = cls.gen_simple_query(
                param=param, field=field, exclude=exclude, groups=groups, orders=orders, limit=size, offset=offset,
                with_del=with_del, with_field=True, db_key=db_key)
            suffix = cls.get_split_suffix(split, param, tz=tz)
            if suffix:
                cls._meta.basetable._table_name, _ = cls.get_sheet_name(suffix)
            if with_count:
                total = await q_set.count()
            data_list = await q_set.values(*out_field)
            return data_list, total
        s_date = param.get(f'{cls._SPLIT_FIELD}') or param.get(f'{cls._SPLIT_FIELD}__gte')
        e_date = param.get(f'{cls._SPLIT_FIELD}__lte')
        if not s_date:
            s_date = tool_dt.cur_dt(tz=tz)
        if not e_date:
            e_date = tool_dt.cur_dt(tz=tz)
        dt_arr = tool_dt.date_range(s_date, e_date, tz=tz)
        dt_arr.reverse()
        count_map, s_count, cur_count = {}, 0, 0
        offset = (page - 1) * size
        for dt in dt_arr:
            s_time, e_time = tool_dt.day_interval(dt, tz=tz)
            new_cond = {**param, **{f'{cls._SPLIT_FIELD}__gte': s_time, f'{cls._SPLIT_FIELD}__lte': e_time}}
            try:
                count = await cls.get_count(new_cond, split=cls._SPLIT_FIELD, tz=tz)
            except Exception as err:
                print(f'GameHistory: {err}')
                count = 0
            if not count:
                continue
            s_count += count
            if (cur_count < size) and (s_count > offset):
                if not cur_count:
                    pos = offset - s_count + count
                else:
                    pos = 0
                limit = (size - cur_count) if ((count - pos) >= size) else count - pos
                count_map[dt] = (limit, pos)
                cur_count += limit
        dlist = []
        for dt, v in count_map.items():
            st, et = tool_dt.day_interval(dt, tz=tz)
            new_cond = {**param, **{f'{cls._SPLIT_FIELD}__gte': st, f'{cls._SPLIT_FIELD}__lte': et}}
            lmt = round(v[0])
            data = await cls.get_by_dict(
                new_cond, field=field, exclude=exclude, groups=groups, orders=orders, with_del=with_del,
                limit=lmt, offset=round(v[1]), db_key=db_key, tz=tz)
            if lmt == 1:
                data = [data]
            data and dlist.extend(data)
        return dlist, s_count

    @classmethod
    async def get_by_pk(
            cls,
            val: Union[int, str],
            field: Union[tuple, list] = None,
            exclude: Union[tuple, list] = None,
            split: Union[str, int, float, datetime, date] = None,
            db_key: str = None,
            tz: str = None):
        """
        通过唯一主键查询
        :param val: 主键值
        :param field: 指定输出的字段
        :param exclude: 指定排除的字段
        :param split: 分表标记 设置分表类型为1时生效，
        :param db_key: 指定数据库配置名
        :param tz: 按日期分表时指定的时区
        """
        if not cls._meta.db_pk_column:
            raise Exception(f'Current sheet {cls.sheet_name()} has no set primary key')
        param = {cls.pk_name(): val}
        query_model, f_list = cls.gen_simple_query(
            param, field=field, exclude=exclude, limit=1, with_field=True, db_key=db_key)
        suffix = cls.get_split_suffix(split, param, tz=tz)
        if suffix:
            cls._meta.basetable._table_name, _ = cls.get_sheet_name(suffix)
        info = await query_model.values(*f_list)
        return info[0] if info else None

    @classmethod
    async def get_by_dict(
            cls,
            param: dict = None,
            field: Union[tuple, list, set] = None,
            exclude: Union[tuple, list, set] = None,
            groups: Union[str, list, tuple] = None,
            orders: Union[str, list, tuple] = '-created',
            split: Union[str, int, float, datetime, date] = None,
            limit=0,
            offset=0,
            with_del=False,
            db_key: str = None,
            tz: str = None):
        query_model, out_filed = cls.gen_simple_query(
            param=param, field=field, exclude=exclude, groups=groups, orders=orders, limit=limit, offset=offset,
            with_del=with_del, with_field=True, db_key=db_key)
        suffix = cls.get_split_suffix(split, param, tz=tz)
        if suffix:
            cls._meta.basetable._table_name, _ = cls.get_sheet_name(suffix)
        dlist = await query_model.values(*out_filed)
        if limit == 1:
            return dlist[0] if dlist else None
        return dlist

    @classmethod
    async def get_count(
            cls,
            param: dict = None,
            with_del=False,
            split: Union[str, int, float, datetime, date] = None,
            db_key: str = None,
            tz: str = None):
        q_set = cls.gen_simple_query(param=param, with_del=with_del, db_key=db_key)
        suffix = cls.get_split_suffix(split, param, tz=tz)
        if suffix:
            cls._meta.basetable._table_name, _ = cls.get_sheet_name(suffix)
        return await q_set.count()

    @classmethod
    async def add_one(cls, param: dict, fun_success=None, db_key: str = None):
        """
        自动分表模式创建
        公共 新增或更新 值为None采用默认值

        :param param: 字典数据模型
        :param fun_success: 执行成功后的补充处理函数 协程函数 参数为生成的Model模型
        :param db_key: 指定数据库配置
        """
        if cls._SPLIT_TYPE:
            raise Exception('分表模型不能使用该方式添加数据')
        cur_time = tool_dt.cur_time()
        param['created'] = cur_time
        ('updated' in cls._meta.db_fields) and param.update({'updated': cur_time})
        new_dict = {field: val for field, val in param.items() if (field in cls._meta.db_fields) and (val is not None)}
        if not new_dict:
            return None
        if (not db_key) or (db_key == 'default'):
            row = await cls.create(**new_dict)
        else:
            db = cls.get_meta_db(db_key)
            row = await cls.create(using_db=db, **new_dict)
        if row:
            callable(fun_success) and await fun_success(row)
            return row
        return None

    @classmethod
    async def update_by_pk(
            cls,
            pk_val: Union[int, str],
            param: dict,
            old_data: dict = None,
            fun_success=None,
            split: Union[str, int, float, datetime, date] = None,
            db_key: str = None,
            tz: str = None):
        if not cls._meta.db_pk_column:
            raise Exception(f'Current sheet {cls.sheet_name()} has no set primary key')
        if not old_data:
            if fun_success:
                raise Exception(f'Must be get [old_data] param for execute the function')
            new_dict = {k: v for k, v in param.items() if (k in cls._meta.db_fields) and (v is not None)}
            split_param = {cls._SPLIT_FIELD: split} if (cls._SPLIT_FIELD not in param) else param
            suffix = cls.get_split_suffix(split, split_param, tz=tz)
        else:
            new_dict = {k: v for k, v in param.items() if (
                    (k in cls._meta.db_fields) and (v is not None)) and (old_data.get(k) != v)}
            split_param = {cls._SPLIT_FIELD: split} if (cls._SPLIT_FIELD not in old_data) else old_data
            suffix = cls.get_split_suffix(split, split_param, tz=tz)
        ('created' in new_dict) and new_dict.pop('created')
        if new_dict:
            ('updated' in cls._meta.db_fields) and new_dict.update({'updated': tool_dt.cur_time()})
            if suffix:
                cls._meta.basetable._table_name, _ = cls.get_sheet_name(suffix)
            if (not db_key) or (db_key == 'default'):
                sta = await cls.filter(**{cls._meta.db_pk_column: pk_val}).limit(1).update(**new_dict)
            else:
                db = cls.get_meta_db(db_key)
                sta = await cls.filter(**{cls._meta.db_pk_column: pk_val}).using_db(db).limit(1).update(**new_dict)
            if sta:
                if callable(fun_success):
                    old_data.update(**new_dict)
                    await fun_success(old_data)
                return new_dict
            return False
        return None

    @classmethod
    async def update_by_cond(
            cls,
            param: dict,
            upinfo: dict,
            limit=1,
            split: Union[str, int, float, datetime, date] = None,
            fun_success=None,
            db_key: str = None,
            tz: str = None):
        new_dict = {k: v for k, v in upinfo.items() if (k in cls._meta.db_fields) and (v is not None)}
        created = ('created' in new_dict) and new_dict.pop('created')
        if new_dict:
            ('updated' in cls._meta.db_fields) and new_dict.update({'updated': tool_dt.cur_time()})
            query_model = cls.gen_simple_query(param, limit=limit, db_key=db_key)
            suffix = cls.get_split_suffix(split, param, tz=tz)
            if suffix:
                cls._meta.basetable._table_name, _ = cls.get_sheet_name(suffix)
            sta = await query_model.update(**new_dict)
            if sta:
                created and new_dict.update({'created': created})
                callable(fun_success) and await fun_success(new_dict)
                return new_dict
            return False
        return None

    @classmethod
    async def del_by_pk(
            cls,
            pk_val: Union[int, str],
            force=False,
            split: Union[str, int, float, datetime, date] = None,
            db_key: str = None,
            tz: str = None,
            fun_success=None,
            fun_ags: Union[list, tuple] = None):
        if not cls._meta.db_pk_column:
            raise Exception(f'Current sheet {cls.sheet_name()} has no set primary key')
        if fun_ags is None:
            fun_ags = ()
        suffix = cls.get_split_suffix(split, {cls._SPLIT_FIELD: split}, tz=tz)
        if suffix:
            cls._meta.basetable._table_name, _ = cls.get_sheet_name(suffix)
        param = {cls._meta.db_pk_column: pk_val}
        if (not hasattr(cls, 'sta_del')) or force:
            sta = await cls.gen_simple_query(param, limit=1, db_key=db_key).delete()
            if sta:
                callable(fun_success) and await fun_success(*fun_ags)
                return True
            return False
        sta = await cls.gen_simple_query(param, limit=1, db_key=db_key).update(sta_del=True)
        if sta:
            callable(fun_success) and await fun_success(*fun_ags)
            return True
        return False

    @classmethod
    async def del_by_cond(
            cls,
            param: dict,
            limit=1,
            force=False,
            split: Union[str, int, float, datetime, date] = None,
            db_key: str = None,
            tz: str = None,
            fun_success=None,
            fun_ags: Union[list, tuple] = None):
        if fun_ags is None:
            fun_ags = ()
        query_model = cls.gen_simple_query(param, limit=limit, with_del=force, with_field=False, db_key=db_key)
        if (not hasattr(cls, 'sta_del')) or force:
            suffix = cls.get_split_suffix(split, param, tz=tz)
            if suffix:
                cls._meta.basetable._table_name, _ = cls.get_sheet_name(suffix)
            sta = await query_model.delete()
            if sta:
                callable(fun_success) and await fun_success(*fun_ags)
                return True
            return False
        suffix = cls.get_split_suffix(split, param, tz=tz)
        if suffix:
            cls._meta.basetable._table_name, _ = cls.get_sheet_name(suffix)
        sta = await query_model.update(sta_del=True)
        if sta:
            callable(fun_success) and await fun_success(*fun_ags)
            return True
        return False

    @classmethod
    async def split_bulk_insert(
            cls,
            data: Union[list[dict], tuple[dict]],
            split: Union[str, int, float, datetime, date] = None,
            use_transaction=False,
            db_key: str = None,
            tz: str = None):
        if not cls._SPLIT_TYPE:
            raise Exception(f'当前数据模型未指定分表模式：{cls.__name__}')
        model_map = {}
        for item in data:
            if not isinstance(item, dict):
                raise Exception(f'检测到无效的数据结构: {item}, {type(item)}')
            (not item.get('created')) and item.update({'created': tool_dt.cur_time()})
            suffix = cls.get_split_suffix(split or cls._SPLIT_FIELD, item, tz=tz)
            model_map[suffix].append(item) if suffix in model_map else model_map.update({suffix: [item]})
        sql_list = []
        for k, v in model_map.items():
            await cls.check_table(k, db_key)
            sql, _, _ = cls.gen_insertup_sql(v, suffix=k)
            sql_list.append(sql)
        try:
            return await cls.exec_insert(sql_list, use_transaction=use_transaction, db_key=db_key)
        except Exception as err:
            raise Exception(f'分表批量写入数据失败：{err}')

    @classmethod
    async def split_add_one(
            cls,
            data: dict,
            split: Union[str, int, float, datetime, date] = None,
            fun_success=None,
            db_key: str = None,
            tz: str = None):
        if not cls._SPLIT_TYPE:
            raise Exception(f'当前数据模型未指定分表模式：{cls.__name__}')
        new_d = {field: val for field, val in data.items() if (field in cls._meta.db_fields) and (val is not None)}
        if not new_d:
            return None
        (not new_d.get('created')) and new_d.update({'created': tool_dt.cur_time()})
        suffix = cls.get_split_suffix(split or cls._SPLIT_FIELD, data, tz=tz)
        try:
            sql, pk_list, cond_list = cls.gen_insertup_sql([new_d], suffix=suffix)
            await cls.check_table(suffix, db_key)
            await cls.exec_sql(sql)
        except Exception as err:
            raise Exception(f'{cls.get_sheet_name(suffix)}分表写入数据失败：{err}')
        if callable(fun_success) and (pk_list or cond_list):
            data = (await fun_success(pk_list)) if pk_list else (await fun_success(cond_list))
            return data
        return 1

    @classmethod
    def gen_insertup_sql(
            cls,
            dlist: Union[list[dict], tuple[dict]],
            upfield: Union[list, tuple] = None,
            add_field: Union[list, tuple] = None,
            mult_field: Union[list, tuple] = None,
            max_field: Union[list, tuple] = None,
            min_field: Union[list, tuple] = None,
            suffix: str = None):
        """
        生成插入数据sql, 使用该写入方式的前提是确保数据库模型不为空的字段有值或存在默认值，否则将无法写入
        """
        if not dlist:
            raise Exception(f'没有可生成sql的数据: {cls.sheet_name(suffix=suffix)}')
        all_field = list(cls._meta.db_fields)
        v_list = []
        for item in dlist:
            if not isinstance(item, dict):
                raise Exception(f'读到不符合规范的数据结构：{item}')
            arr = [cls.check_val(item.get(k), k, cls) for k in all_field]
            v_list.append("({})".format(','.join(arr)))
        sql = F"INSERT INTO `{cls.sheet_name(suffix=suffix)}`({','.join([f'`{f}`' for f in all_field])}) VALUES "
        sql += ",".join(v_list)
        pk_list = []
        cond_list = []
        if not any((upfield, add_field, mult_field, max_field, min_field)):
            return sql, pk_list, cond_list
        for item in dlist:
            pk_val = item.get(cls.pk_name())
            if pk_val:
                pk_list.append(pk_val)
                continue
            if not cls._meta.unique_together:
                raise Exception('数据结构错误,更新必须指定数据模型的唯一键或唯一索引')
            cond = {k: item.get(k) for k in cls._meta.unique_together[0]}
            if not cond:
                raise Exception('数据结构错误,更新必须指定包含数据唯一索引的值')
            cond_list.append(cond)
        if cls._meta.db.schema_generator.DIALECT == 'mysql':
            sql += "AS up_rows ON DUPLICATE KEY UPDATE "
            if upfield:
                sql += f'{",".join(["{0}=up_rows.{0}".format(field) for field in upfield if field in cls._meta.db_fields])},'
            if add_field:
                sql += f'{",".join(["{0}={0}+up_rows.{0}".format(field) for field in add_field if field in cls._meta.db_fields])},'
            if mult_field:
                sql += f'{",".join(["{0}={0}*up_rows.{0}".format(field) for field in mult_field if field in cls._meta.db_fields])},'
            if max_field:
                sql += f'{",".join(["{0}=GREATEST({0},VALUES({0}))".format(field) for field in max_field if field in cls._meta.db_fields])},'
            if min_field:
                sql += f'{",".join(["{0}=LEAST({0},VALUES({0}))".format(field) for field in min_field if field in cls._meta.db_fields])},'
            sql = sql[:-1]
        elif cls._meta.db.schema_generator.DIALECT == 'postgres':
            if pk_list and cond_list:
                raise Exception('pgsql更新模式下所有数据只能指定唯一键或唯一索引两者之一的方式')
            if pk_list:
                up_keys = cls.pk_name()
            else:
                up_keys = ','.join(list(cls._meta.unique_together[0]))
            sql += f"ON CONFLICT ({up_keys}) DO UPDATE SET "
            if upfield:
                sql += ','.join([f"{filed}=EXCLUDED.{filed}" for filed in upfield])
        return sql, pk_list, cond_list

    @classmethod
    def gen_insertup_for_old_mysql(cls,
            dlist: Union[list[dict], tuple[dict]],
            upfield: Union[list, tuple] = None,
            add_field: Union[list, tuple] = None,
            mult_field: Union[list, tuple] = None,
            max_field: Union[list, tuple] = None,
            min_field: Union[list, tuple] = None,
            suffix: str = None):
        """
        生成插入数据sql, 使用该写入方式的前提是确保数据库模型不为空的字段有值或存在默认值，否则将无法写入
        """
        if cls._meta.db.schema_generator.DIALECT != 'mysql':
            raise Exception(f'该方式只针对于mysql: {cls.sheet_name(suffix=suffix)}')
        if not dlist:
            raise Exception(f'没有可生成sql的数据: {cls.sheet_name(suffix=suffix)}')
        all_field = list(cls._meta.db_fields)
        v_list = []
        for item in dlist:
            if not isinstance(item, dict):
                raise Exception(f'读到不符合规范的数据结构：{item}')
            arr = [cls.check_val(item.get(k), k, cls) for k in all_field]
            v_list.append("({})".format(','.join(arr)))
        sql = F"INSERT INTO `{cls.sheet_name(suffix=suffix)}`({','.join([f'`{f}`' for f in all_field])}) VALUES "
        sql += ",".join(v_list)
        pk_list = []
        cond_list = []
        if not any((upfield, add_field, mult_field, max_field, min_field)):
            return sql, pk_list, cond_list
        for item in dlist:
            pk_val = item.get(cls.pk_name())
            if pk_val:
                pk_list.append(pk_val)
                continue
            if not cls._meta.unique_together:
                raise Exception('数据结构错误,更新必须指定数据模型的唯一键或唯一索引')
            cond = {k: item.get(k) for k in cls._meta.unique_together[0]}
            if not cond:
                raise Exception('数据结构错误,更新必须指定包含数据唯一索引的值')
            cond_list.append(cond)
        sql += " ON DUPLICATE KEY UPDATE "
        if upfield:
            sql += f'{",".join(["{0}=VALUES({0})".format(field) for field in upfield if field in cls._meta.db_fields])},'
        if add_field:
            sql += f'{",".join(["{0}={0}+VALUES({0})".format(field) for field in add_field if field in cls._meta.db_fields])},'
        if mult_field:
            sql += f'{",".join(["{0}={0}*VALUES({0})".format(field) for field in mult_field if field in cls._meta.db_fields])},'
        if max_field:
            sql += f'{",".join(["{0}=GREATEST({0},VALUES({0}))".format(field) for field in max_field if field in cls._meta.db_fields])},'
        if min_field:
            sql += f'{",".join(["{0}=LEAST({0},VALUES({0}))".format(field) for field in min_field if field in cls._meta.db_fields])},'
        sql = sql[:-1]
        return sql, pk_list, cond_list



import os

from sanic.config import Config

from nsanic.libs.consts import GLOBAL_TZ


class BaseConf(Config):

    SERVER_NAME = 'MAIN'
    '''服务名称'''
    SERVER_ID = 'M0001'
    '''服务ID，建议4-6个字符 不同的服务之间该标识定义必须不一致'''
    RUN_PORT = 8800
    '''服务监听端口，不同的服务请使用不同的监听端口，否则无法启动'''
    RUN_WORKER = 1
    '''工作进程，该项配置可依据CPU核心数配置，最佳值为CPU核心数'''
    RUN_FAST = True
    HOST = '127.0.0.1'
    '''监听主机'''
    VER_CODE = None
    '''接口默认版本号'''
    DEBUG_MODE = False
    '''是否开启debug模式'''
    ACCESS_LOG = False
    '''是否开启access日志'''
    CHARSET = 'utf-8'
    '''全局字符集'''
    RESP_TYPE = 'JSON'
    USE_TZ = False
    '''数据库时区转化'''
    TIME_ZONE = GLOBAL_TZ
    '''时区配置'''
    REQUEST_TIMEOUT = 60
    '''	请求超时时间'''
    RESPONSE_TIMEOUT = 60
    '''响应超时时间'''
    FALLBACK_ERROR_FORMAT = 'json'
    '''错误响应格式'''
    FORWARDED_FOR_HEADER = 'X-Forwarded-For'
    '''代理IP或客户端IP请求头配置'''
    FORWARDED_SECRET = None
    '''用于安全识别特定的代理服务器'''
    GRACEFUL_SHUTDOWN_TIMEOUT = 30.0
    '''强制关闭非空闲连接的等待时间(秒)'''
    KEEP_ALIVE = True
    '''是否启用长连接'''
    KEEP_ALIVE_TIMEOUT = 30
    '''长连接超时时间'''
    REAL_IP_HEADER = 'X-Real-IP'
    '''客户端真实IP'''
    REGISTER = True
    '''是否应用程序注册'''
    REQUEST_BUFFER_QUEUE_SIZE = 10000
    '''请求缓存区队列大小'''
    REQUEST_ID_HEADER = 'X-Request-ID'
    '''请求头中的请求ID名称'''
    REQUEST_MAX_SIZE = 100000000
    '''Request 的最大字节数'''
    WEBSOCKET_MAX_SIZE = 2 ^ 20
    '''websocket 传入消息最大字节数'''
    WEBSOCKET_PING_INTERVAL = 30
    '''websocket ping 帧 发送间隔'''
    WEBSOCKET_PING_TIMEOUT = 20
    '''websocket pong 帧 响应超时时间'''

    CORS_AUTOMATIC_OPTIONS = True

    # 跨域访问允许的域名 * -- 所有域名
    ALLOW_ORIGIN = ['localhost']
    '''所允许的域'''
    ALLOW_METHOD = ['GET', 'POST', 'PUT', 'DELETE', 'OPTIONS']
    '''允许的请求方法'''
    ALLOW_HEADER = ['Authorization']
    '''允许的请求头'''
    ALLOW_CREDENTIALS = True
    '''是否允许携带cookie类验证信息'''
    ORIGIN_MAX_AGE = 300
    '''预检最大有效期'''

    PROCESSING_REQ = 'PROCESSING_REQUEST'
    '''当前正在处理的请求缓存KEY'''

    CHANNEL_SYSTEM = 'CHANNEL_SYSTEM_MSG'
    '''统一系统消息频道'''

    # 日志基础路径
    LOG_PATH = os.path.join(os.getcwd(), 'logs')

    MODEL_LIST = ['main']
    '''数据库迁移基础表 更换配置请重写'''
    MODEL_EXTRA = ['extra']
    '''只用模型而不需要在当前项目迁移的表，请配置到该目录'''
    proc_name = ''

    # 默认Redis缓存配置 更换配置请重写
    CONF_RDS = {
        'default': {
            'host': '',
            'port': 6379,
            'password': '',
            'db': 1,
            # 'max_connections': 5,  # 最大连接数
            # 'decode_responses': True,  # 默认解析字符串，不解析默认为bytes类型
            'socket_connect_timeout': 5,  # 连接超时时长 s
        }
    }

    # tortoise-orm 默认数据库配置 更换配置请重写
    CONF_DB = {
        'default': {
            'engine': 'tortoise.backends.asyncpg',
            'credentials': {
                'host': '',
                'port': '5432',
                'user': '',
                'password': '',
                'database': '',
            },
            'options': {
                'charset': 'utf8mb4',
                'minsize': 5,
                'maxsize': 20,
            }
        }
    }

    @classmethod
    def makeup_db_conf(cls, model_list: list):
        """自定义分离式多库模式请重写该方法"""
        db_conf = {
            'apps': {cls.SERVER_NAME: {'models': model_list}},
            'connections': cls.CONF_DB,
            'use_tz': cls.USE_TZ if cls.TIME_ZONE else False
        }
        cls.USE_TZ and db_conf.update({'timezone': cls.TIME_ZONE})
        return db_conf

    @classmethod
    def db_conf(cls):
        """数据库配置"""
        models = cls.MODEL_LIST + cls.MODEL_EXTRA
        model_list = [f'{cls.SERVER_NAME}.model_db.{item}' for item in models]
        return cls.makeup_db_conf(model_list) if cls.CONF_DB else None

    @classmethod
    def migrate_db(cls):
        model_list = [f'{cls.SERVER_NAME}.model_db.{item}' for item in cls.MODEL_LIST]
        model_list.insert(0, 'aerich.models')
        return cls.makeup_db_conf(model_list)

    @classmethod
    def log_conf(cls):
        (not os.path.exists(cls.LOG_PATH)) and os.mkdir(cls.LOG_PATH)
        log_path = os.path.join(cls.LOG_PATH, cls.SERVER_NAME)
        (not os.path.exists(log_path)) and os.mkdir(log_path)
        mode = 'DEBUG' if cls.DEBUG_MODE else 'INFO'
        return {
            'version': 1,
            'disable_existing_loggers': False,
            'handlers': {
                'console': {
                    'class': 'logging.handlers.RotatingFileHandler',
                    'filename': os.path.join(log_path, 'run.log'),
                    'level': mode,
                    'formatter': 'generic',
                    'maxBytes': 1024 * 1024 * 200,
                    'backupCount': 5,
                    'encoding': 'utf-8'
                },
                'access': {
                    'class': 'logging.handlers.RotatingFileHandler',
                    'level': mode,
                    'formatter': 'access',
                    'filename': os.path.join(log_path, 'access.log'),
                    'maxBytes': 1024 * 1024 * 200,
                    'backupCount': 5,
                    'encoding': 'utf-8'
                },
                'error': {
                    'class': 'logging.handlers.RotatingFileHandler',
                    'level': 'ERROR',
                    'formatter': 'generic',
                    'filename': os.path.join(log_path, 'error.log'),
                    'maxBytes': 1024 * 1024 * 200,
                    'backupCount': 5,
                    'encoding': 'utf-8'
                },
            },
            'formatters': {
                "generic": {
                    "format": "%(asctime)s [%(process)s] [%(levelname)s] %(message)s",
                    "datefmt": "[%Y-%m-%d %H:%M:%S %z]",
                },
                "access": {
                    "format": "%(asctime)s - (%(name)s)[%(levelname)s][%(host)s]: "
                    + "%(request)s %(message)s %(status)s %(byte)s",
                    "datefmt": "[%Y-%m-%d %H:%M:%S %z]",
                },
            },
            'loggers': {
                "sanic.root": {"level": "INFO", "handlers": ["console"]},
                "sanic.error": {
                    "handlers": ["error"],
                    "propagate": True,
                    "qualname": "sanic.error",
                },
                "sanic.access": {
                    "handlers": ["access"],
                    "propagate": True,
                    "qualname": "sanic.access",
                },
                "sanic.server": {
                    "handlers": ["console"],
                    "propagate": True,
                    "qualname": "sanic.server",
                },
            }
        }



import logging
import multiprocessing
import os
from logging import handlers


class NLogger:
    log_fmt = "%(asctime)s - %(process)d:%(lineno)s:%(levelname)s - %(message)s"
    '''日志格式'''
    date_fmt = '%Y-%m-%d %H:%M:%S'
    '''日期格式'''
    base_path = os.path.join(os.getcwd(), 'logs')
    '''基础路径'''
    full_path = None
    proc_split = 0
    '''进程隔离模式 0-不隔离 1-按进程标识隔离 2-按进程pid隔离'''
    proc_tab = ""
    '''进程标识  proc_split 为1时有效'''
    log_split = 1
    '''日志分割模式  0-单日志 1-按固定大小分割的滚动日志 2-按时间分割的滚动日志'''
    t_when = 'D'
    '''按时间滚动方式 log_split为2有效 可取值S-秒 M-分 H-小时 D-天 W-周'''
    interval = 1
    '''按时间滚动方式间隔周期 log_split为2有效'''
    keep_count = 10
    '''针对滚动日志类型或时间分割类型日志需要保留的日志个数 log_split为1或2有效'''
    log_size = 104857600
    '''滚动类型日志单个日志的最大限制大小 log_split为1有效'''
    encoding = 'utf-8'
    '''日志存储的编码格式'''
    level = 'INFO'
    '''日志默认记录级别'''
    console = True
    '''控制台输出'''

    @classmethod
    def init_conf(cls, base_path: str = None, folder: str = None, log_fmt: str = None, dt_fmt: str = None,
                  proc_split: int = 0, proc_tab: str = None, log_split: int = 0, t_when: str = None, interval=1,
                  keeps: int = None, log_size: int = None, encoding: str = None, level: str = None, console=True):
        """
        初始化日志管理器
        :param base_path: 日志基础路径
        :param folder: 路径下的存储目录
        :param log_fmt: 日志格式
        :param dt_fmt: 日期格式
        :param proc_split: 进程隔离模式 0-不隔离 1-按进程标识隔离 2-按进程pid隔离
        :param proc_tab: 进程标识  proc_split 为1时有效
        :param log_split: 日志分割模式  0-单日志 1-按固定大小分割的滚动日志 2-按时间分割的滚动日志
        :param t_when: 按时间滚动方式设置 可取值S-秒 M-分 H-小时 D-天 W-周
        :param interval: 按时间滚动方式间隔周期 log_split为2有效
        :param keeps: 针对滚动日志类型或时间分割类型日志需要保留的日志个数 log_split为1或2有效
        :param log_size: 滚动类型日志单个日志的最大限制大小 log_split为1有效
        :param encoding: 日志存储的编码格式 默认utf-8
        :param level: 日志默认记录级别
        :param console: 加控制台输出
        """
        if base_path:
            cls.base_path = base_path
        (not os.path.exists(cls.base_path)) and os.mkdir(cls.base_path)
        if folder:
            cls.full_path = os.path.join(os.path.join(cls.base_path, folder))
            (not os.path.exists(cls.full_path)) and os.mkdir(cls.full_path)
        else:
            cls.full_path = cls.base_path
        if log_fmt:
            cls.log_fmt = log_fmt
        if dt_fmt:
            cls.date_fmt = dt_fmt
        if proc_split and proc_split in (0, 1, 2):
            cls.proc_split = proc_split
        if cls.proc_split == 1:
            cls.proc_tab = f"_{proc_tab.lower()}" if proc_tab else f"_{multiprocessing.current_process().name.lower()}"
        elif cls.proc_split == 2:
            cls.proc_tab = f"_{os.getpid()}"
        if log_split:
            cls.log_split = log_split
        if cls.log_split == 2:
            cls.t_when = t_when if t_when else 'D'
            cls.interval = interval if interval else 1
            cls.keep_count = keeps if keeps else 10
        elif cls.log_split == 1:
            cls.log_size = log_size if log_size else 104857600
            cls.keep_count = keeps if keeps else 10
        if encoding:
            cls.encoding = encoding
        if level:
            cls.level = level
        cls.console = console

    @classmethod
    def _mklog(cls, log_name='runlog'):
        f_name = f"{log_name}{cls.proc_tab}.log"
        fp = os.path.join(cls.full_path, f_name)
        log = logging.getLogger(fp)
        if log.handlers:
            return log
        formatter = logging.Formatter(fmt=cls.log_fmt, datefmt=cls.date_fmt)
        if cls.log_split == 1:
            log_handler = handlers.RotatingFileHandler(
                filename=fp, mode="a", maxBytes=cls.log_size, backupCount=cls.keep_count, encoding=cls.encoding)
        elif cls.log_split == 2:
            log_handler = handlers.TimedRotatingFileHandler(
                filename=fp, when=cls.t_when, interval=cls.interval, backupCount=cls.keep_count, encoding=cls.encoding)
        else:
            log_handler = handlers.RotatingFileHandler(filename=fp, mode="a", encoding=cls.encoding)
        log.setLevel(cls.level)
        log_handler.setFormatter(formatter)
        log.addHandler(log_handler)
        if cls.console:
            ch = logging.StreamHandler()
            ch.setLevel(logging.DEBUG)
            log.addHandler(ch)
        return log

    @classmethod
    def error(cls, *out_info):
        """代码或执行失败或报错的日志 调用该方法前请确认init_conf已调用"""
        log = cls._mklog(log_name='error')
        log.error(str(out_info[0]) if len(out_info) == 1 else str(out_info))

    @classmethod
    def critical(cls, *out_info):
        log = cls._mklog(log_name='error')
        log.critical(str(out_info[0]) if len(out_info) == 1 else str(out_info))

    @classmethod
    def info(cls, *out_info):
        """标记或打印类的日志 调用该方法前请确认init_conf已调用"""
        log = cls._mklog()
        log.info(str(out_info[0]) if len(out_info) == 1 else str(out_info))

    @classmethod
    def warning(cls, *out_info):
        """警告类日志 调用该方法前请确认init_conf已调用"""
        log = cls._mklog()
        log.warning(str(out_info[0]) if len(out_info) == 1 else str(out_info))

    @classmethod
    def debug(cls, *out_info):
        """debug日志 调用该方法前请确认init_conf已调用"""
        log = cls._mklog()
        log.debug(str(out_info[0]) if len(out_info) == 1 else str(out_info))



import base64
from Crypto import Random
from Crypto.Cipher import PKCS1_v1_5
from Crypto.PublicKey import RSA


def rsa_decode(item_str: str, pri_key: str, log_fun=None):
    """rsa私钥加密"""
    if not isinstance(item_str, str):
        item_str = str(item_str)
    cipher_pri = PKCS1_v1_5.new(RSA.importKey(pri_key))
    try:
        decp_str = cipher_pri.decrypt(base64.b64decode(item_str.encode('utf-8')), Random.new().read)
    except Exception as err:
        err_info = f"RSA解析失败:{err}\n原数据:{item_str}"
        log_fun(err_info) if callable(log_fun) else print(err_info)
        return None
    return decp_str.decode()


def rsa_encode(item_str: str, pub_key: str, log_fun=None):
    """rsa公钥加密"""
    if not isinstance(item_str, str):
        item_str = str(item_str)
    try:
        cipher_pub = PKCS1_v1_5.new(RSA.importKey(pub_key))
        item_ecp = cipher_pub.encrypt(item_str.encode('utf-8'))
        return base64.b64encode(item_ecp).decode()
    except Exception as err:
        err_info = f'加密数据出错,源数据{item_str}\n错误信息:{err}'
        log_fun(err_info) if callable(log_fun) else print(err_info)


def create_rsa(df_len=2048):
    """
    创建RSA密钥对

    :param df_len: 默认允许加密长度2048--最大加密245长度  设置128生成时将会更快速--最多只能加密117长度的字符串
    :return 私钥,公钥
    """
    gen = RSA.generate(df_len)
    return (gen.exportKey('PEM')).decode('utf8'), (gen.publickey().exportKey()).decode('utf8')



from typing import Union

from nsanic.libs.consts import Code, StaCode
from nsanic.libs.tool import json_encode, json_parse


def pack_msg(
        cmd_type: Union[int, str],
        cmd_code: Union[int, str],
        data: Union[dict, list] = None,
        code: Code = StaCode.PASS,
        hint='',
        req=None,
        log_fun=None):
    """
    WS打包消息
    :param cmd_type: 消息命令类型
    :param cmd_code: 消息标识位
    :param data: 消息体数据
    :param code: 响应状态码
    :param hint: 提示信息
    :param req: 客户端消息标记
    :param log_fun: 日志记录方法
    """
    cmd = None
    if isinstance(cmd_code, int) and isinstance(cmd_type, int):
        if 0 <= cmd_code < 1000:
            cmd = cmd_type * 1000 + cmd_code
    if isinstance(cmd_code, str) or isinstance(cmd_type, str):
        cmd = f'{cmd_type}.{cmd_code}'
    if cmd:
        body = {'cmd': cmd, 'data': data, 'code': code.val, 'req': req, 'hint': hint or code.msg}
        return json_encode(body, u_byte=True, log_fun=log_fun)
    err_info = f'打包消息出错,消息结构错误: {cmd_type}, {cmd_code},{data},{code.val},{hint or code.msg},{req}'
    log_fun(err_info) if callable(log_fun) else print(err_info)
    return None


def parse_msg(data, log_fun=None):
    """消息解析"""
    msg = json_parse(data, log_fun=log_fun)
    if (not msg) or (not isinstance(msg, dict)):
        return None, None, None, None
    try:
        cmd = msg.get('cmd')
        if isinstance(cmd, int):
            cmd_type, cmd_code = cmd // 1000, cmd % 1000
        else:
            cmd_type, cmd_code = str(cmd).split('.')
    except Exception as err:
        log_fun(f'{data}消息解析出错：{err}') if log_fun else print(f'{data}消息解析出错：{err}')
        return None, None, None, None
    return cmd_type, cmd_code, json_parse(msg.get('data')), msg.get('req')



import asyncio
import random
import re
import string
import time

from nsanic.libs.tool_dt import cur_dt


class RngMaker:
    """随机数生成模型"""
    _STR_LIST = set()
    _PRE_STR = ''
    _PRE_NUM = ''
    _DEF_LEN = 32
    _PRE_LEN = 0
    _PRE_NUM_LEN = 0

    @classmethod
    def init(cls, pre_str='', default_len=32):
        cls._PRE_STR = pre_str
        cls._DEF_LEN = default_len
        cls._PRE_NUM = "".join([num for num in re.findall(r'\d+',  cls._PRE_STR)])
        cls._PRE_LEN = len(cls._PRE_STR)
        cls._PRE_NUM_LEN = len(cls._PRE_NUM)

    @staticmethod
    def __date_str(fmt='%Y%m%d%H%M%S%f'):
        """日期时间模式"""
        return cur_dt().strftime(fmt)

    @staticmethod
    def __tamp_int():
        """时间戳毫秒秒模式"""
        return int(time.time() * 1000)

    @staticmethod
    def __hex_num(num: int):
        """转16进制"""
        return hex(num)[2:]

    @classmethod
    def __lower_hex(cls, str_len: int, is_stamp=True):
        """全小写字母随机"""
        hstr = cls.__hex_num(cls.__tamp_int() if is_stamp else int(cls.__date_str()))
        _len = str_len - len(hstr)
        if _len > cls._PRE_LEN:
            _len -= cls._PRE_LEN
        return f"{cls._PRE_STR.lower()}{hstr}{''.join(random.sample(string.ascii_lowercase + string.digits, _len))}"

    @classmethod
    def __letters_hex(cls, str_len: int, is_stamp=True):
        """全字母随机"""
        h_str = cls.__hex_num(cls.__tamp_int() if is_stamp else int(cls.__date_str()))
        length = str_len - len(h_str)
        if length > cls._PRE_LEN:
            length -= cls._PRE_LEN
        return f"{cls._PRE_STR}{h_str}{''.join(random.sample(string.ascii_letters + string.digits, length))}"

    @classmethod
    def __all_num(cls, str_len: int, is_stamp=True):
        """纯数字随机"""
        gen_str = str(cls.__tamp_int()) if is_stamp else cls.__date_str()
        length = str_len-len(gen_str)
        if length > cls._PRE_NUM_LEN:
            length -= cls._PRE_NUM_LEN
        return f"{cls._PRE_NUM}{gen_str}{''.join(random.sample(string.digits, length))}"

    @classmethod
    async def __creator(cls, str_len: int, create_random, is_stamp=True) -> str:
        while True:
            if len(cls._STR_LIST) > 20000:
                await asyncio.sleep(0.002)
                cls._STR_LIST.clear()
            r_str = create_random(str_len, is_stamp)
            if r_str not in cls._STR_LIST:
                cls._STR_LIST.add(r_str)
                return r_str

    @classmethod
    async def gen_str(cls, str_len=30, short_mode=True):
        """
        (不重复)按时间生成指定长度不重复的字符串，默认长度30，短模式最短15 最长46 非短模式最短21 最长50 不在该的长度范围取默认值

        :param str_len: 生成字符串长度. 默认长度30
        :param short_mode: 是否使用短模式 默认 是
        """
        if short_mode and (15 <= str_len <= 46):
            return await cls.__creator(str_len, cls.__lower_hex, short_mode)
        if 21 <= str_len <= 50:
            return await cls.__creator(str_len, cls.__lower_hex, False)
        return await cls.__creator(cls._DEF_LEN, cls.__lower_hex, short_mode)

    @classmethod
    async def gen_num(cls, str_len=18, is_stamp=True):
        """
        (不重复)按时间生成指定长度不重复的数字单号，默认长度18，时间戳模式最短18 最长26 非时间戳模式最短18 最长32 不在该的长度范围取默认值

        :param str_len: 生成字符串长度. 默认长度18
        :param is_stamp: 是否使用时间戳模式 默认 是
        """
        if is_stamp and (18 <= str_len <= 26):
            return await cls.__creator(str_len, cls.__all_num, is_stamp)
        if 18 <= str_len <= 32:
            return await cls.__creator(str_len, cls.__all_num, False)
        return await cls.__creator(18, cls.__all_num, is_stamp)

    @staticmethod
    def mk_int(min_num: int, max_num: int, ji_shu=0):
        """
        (允许重复)按最大值最小值范围取随机数

        :param min_num: 最小值
        :param max_num: 最大值
        :param ji_shu: 取值基数 该参数不为0时 取到的值会对对该值取整再乘以该值
        """
        res = random.randint(min_num, max_num)
        if ji_shu:
            return (res // ji_shu) * ji_shu
        return res

    @staticmethod
    def mk_num(length=5):
        """(允许重复)取指定数量级范围内的随机数"""
        return random.randint(10 ** (length - 1), 10 ** length - 1)

    @staticmethod
    def mk_str(length: int = 16, use_caps=False):
        """随机字符串 默认16位"""
        if use_caps:
            return ''.join(random.sample(string.ascii_lowercase + string.ascii_uppercase + string.digits, length))
        return ''.join(random.sample(string.ascii_lowercase + string.digits, length))

    @staticmethod
    def mk_arr_one(item_list: (list, tuple), weights: (list, tuple) = None):
        """list随机筛选某项, 带权重的话权重配置长度必须等于列表或元组长度"""
        if item_list and isinstance(item_list, list or tuple):
            if weights and (len(item_list) != len(weights)):
                return None
            return random.choices(item_list, weights=weights)
        return None



import asyncio
import traceback
from typing import Union

from sanic import Request
from sanic.exceptions import WebsocketClosed
from sanic.server.protocols.websocket_protocol import WebSocketProtocol
from websockets.protocol import State

from nsanic.libs import tool_ws
from nsanic.libs.component import ConfMeta
from nsanic.libs.consts import StaCode
from nsanic.libs.manager import WsConnector


class BaseWebsocket(ConfMeta):
    STA_CODE = StaCode
    CMD_MAP = {}
    '''服务命令映射'''
    ws_manager = WsConnector

    dft_type: Union[int, str] = 1
    '''系统默认消息类型'''
    beat_code: Union[int, str] = 1
    '''心跳消息标码'''
    reject_code: Union[int, str] = 2
    '''弹回消息标码'''
    delay_fail = 10
    '''授权或无效连接响应延时（秒）'''

    def __init__(self):
        self.fun_pack_msg: callable = tool_ws.pack_msg
        self.fun_parse_msg: callable = tool_ws.parse_msg
        self.delay_pre = round(self.delay_fail * 0.6, 2)
        self.delay_aft = round(self.delay_fail * 0.4, 2)
        for v in self.CMD_MAP.values():
            v.STA_CODE = self.STA_CODE

    @classmethod
    def init(cls):
        return cls()

    @staticmethod
    def get_ukey(uinfo):
        return uinfo.get('uid') or uinfo.get('id') or uinfo.get('userid') if isinstance(uinfo, dict) else uinfo

    async def init_offline_queue(self):
        while 1:
            off_ws = await self.ws_manager.OFFLINE_QUEUE.get()
            if off_ws:
                ukey = getattr(off_ws, 'ukey', None)
                if not ukey:
                    continue
                cur_ws = self.ws_manager.get_ws(ukey)
                if cur_ws and (cur_ws.ws_proto.id == off_ws.ws_proto.id):
                    await self.on_offline(ukey)
                for k, v in self.CMD_MAP.items():
                    try:
                        await v.offline(ukey)
                    except Exception as err:
                        self.logerr(f'用户掉线处理出错：{ukey}, {err}')

    async def client_listen(self, ws, uinfo):
        ukey = self.get_ukey(uinfo)
        while ws.ws_proto.state == State.OPEN:
            cmd, fcmd, data, extra = self.fun_parse_msg(await ws.recv(), log_fun=self.logerr)
            if cmd and (cmd == self.dft_type) and (fcmd == self.beat_code):
                await ws.send(self.fun_pack_msg(
                    self.dft_type, self.beat_code, data=data, code=self.STA_CODE.PASS, req=extra))
                continue
            fun = self.CMD_MAP.get(cmd)
            if fun and callable(fun.funapi):
                mlist, is_end, in_send = [], False, False
                try:
                    msg_arg = await fun.funapi(fcmd, uinfo, data)
                    is_end = msg_arg.pop(0)
                    is_mult = msg_arg.pop(0)
                    if not is_mult:
                        mlist = [self.fun_pack_msg(cmd, *msg_arg, req=extra)]
                    else:
                        if isinstance(msg_arg[1], dict):
                            mlist = [self.fun_pack_msg(cmd, c, d, msg_arg[2], req=extra) for c, d in msg_arg[1].items()]
                        elif isinstance(msg_arg[1], list):
                            mlist = [self.fun_pack_msg(cmd, msg_arg[0], d, msg_arg[2], req=extra) for d in msg_arg[1]]
                    in_send = True
                    for msg in mlist:
                        msg and (await ws.send(msg))
                    if is_end:
                        return await self.on_offline(ukey)
                except WebsocketClosed:
                    if (not is_end) and mlist and hasattr(self.ws_manager, 'save_to_histories'):
                        await self.ws_manager.save_to_histories(ukey, mlist)
                    return await self.on_offline(ukey)
                except Exception as err:
                    self.logerr(f'{err}:{traceback.format_exc()}')
                    if (not is_end) and mlist and in_send and hasattr(self.ws_manager, 'save_to_histories'):
                        await self.ws_manager.save_to_histories(ukey, mlist)
                    await ws.send(self.fun_pack_msg(
                        cmd, fcmd, code=self.STA_CODE.FAIL, req=extra, hint='A error cause failed'))
            else:
                await ws.send(self.fun_pack_msg(
                    self.dft_type, self.reject_code, code=self.STA_CODE.FAIL, hint='Unspecified message'))

    async def wsrouter(self, req: Request, ws):
        """Websocket路由服务"""
        uinfo = await self.wsauth(req, ws)
        if not uinfo:
            self.delay_aft and (await asyncio.sleep(self.delay_aft))
            return await ws.close()
        key_info = await self.check_ws_conn(uinfo, req, ws)
        if not key_info:
            self.delay_aft and (await asyncio.sleep(self.delay_aft))
            return await ws.close()
        await self.update_status(uinfo)
        ukey = self.get_ukey(uinfo)
        await self.ws_manager.send_histories(ukey)
        setattr(ws, 'ukey', ukey)
        setattr(ws, 'connector_name', self.ws_manager.__name__)
        isinstance(key_info, str) and (await self.ws_manager.set_ws(ukey, ws, key_info=key_info))
        await self.client_listen(ws, uinfo)

    async def check_ws_conn(self, uinfo, req, ws) -> Union[bool, str, None]:
        timestamp = req.headers.get('Timestamp') or req.args.get('Timestamp')
        if not timestamp:
            self.delay_pre and (await asyncio.sleep(self.delay_pre))
            await ws.send(self.fun_pack_msg(
                self.dft_type, self.reject_code, code=self.STA_CODE.FAIL, hint='missing params,reject'))
            return None
        ukey = self.get_ukey(uinfo)
        timestamp = int(timestamp)
        if hasattr(self.ws_manager, 'WS_ONLINE_KEY'):
            old_info = await self.ws_manager.check_old_ws(ukey)
            if old_info:
                old_arr = old_info.split(b'--' if isinstance(old_info, bytes) else '--')
                if int(old_arr[1]) >= timestamp:
                    self.delay_pre and (await asyncio.sleep(self.delay_pre))
                    await ws.send(self.fun_pack_msg(
                        self.dft_type, self.reject_code, code=self.STA_CODE.FAIL, hint='invalid connection,reject'))
                    return None
        key_info = None
        if hasattr(self.ws_manager, 'MAIN_CHANNEL'):
            key_info = f"{self.ws_manager.MAIN_CHANNEL}--{timestamp}"
        return key_info or True

    async def update_status(self, uinfo):
        """更新用户监听的频道信息 更新连接状态等相关内容写在这里"""
        pass

    async def wsauth(self, req, ws) -> (object, dict, str, int):
        """基础接口认证处理 接入/认证等"""
        print(ws, req)
        raise Exception('接口未指定授权，不允许访问')

    async def on_offline(self, ukey):
        return await self.ws_manager.offline(ukey)


class WsProtocol(WebSocketProtocol):
    connector_map = {}

    def connection_lost(self, exc):
        async def to_queue(ws, conn_name):
            connector = self.connector_map.get(conn_name)
            connector and (await connector.OFFLINE_QUEUE.put(ws))

        super(WsProtocol, self).connection_lost(exc)
        if self.websocket is not None:
            u_key, name = getattr(self.websocket, 'ukey', None), getattr(self.websocket, 'connector_name', None)
            if (not u_key) or (not name):
                return
            loop = asyncio.get_running_loop()
            loop.create_task(to_queue(self.websocket, name))



import asyncio
import multiprocessing
from asyncio import AbstractEventLoop, Protocol
from socket import socket
from ssl import SSLContext
from typing import Optional, Union

from sanic import Sanic
from sanic.http.constants import HTTP
from sanic.mixins.startup import HTTPVersion

from nsanic.base_conf import BaseConf
from nsanic.libs import consts
from nsanic.libs.mk_random import RngMaker
from nsanic.libs.mult_log import NLogger
from nsanic.libs.rds_client import RdsClient


class InitServer:

    def __init__(
            self,
            conf,
            middlewares: (list, tuple) = None,
            bp_arr: (list, tuple) = None,
            exceptions: (list, tuple) = None,
            forbid_before_evt=True,
            start_evt: list = None,
            stop_evt: list = None):
        """
        :param conf: BaseConf 对象
        :param middlewares: middleware 对象列表或元组
        :param bp_arr: 蓝图元组或集合
        :param exceptions: 全局异常捕获器
        :param forbid_before_evt: 是否在启动事件完成前进行接口拦截
        :param start_evt: 启动需要注册的事件 可选参数 **_
        :param stop_evt: 服务结束前的注册事件 可选参数 **_
        """
        self.__conf: BaseConf = conf
        consts.GLOBAL_TZ = conf.TIME_ZONE
        RngMaker.init(pre_str=self.__conf.SERVER_ID)
        name = multiprocessing.current_process().name
        name_arr = name.split('-')
        proc_name = f"{name_arr[-2]}{name_arr[-1]}" if len(name_arr) >= 2 else name
        self.proc_name = proc_name
        NLogger.init_conf(base_path=self.__conf.LOG_PATH, folder=self.__conf.SERVER_NAME.lower(), proc_split=1, proc_tab=proc_name, console=conf.DEBUG_MODE)
        if not forbid_before_evt:
            consts.GLOBAL_SRV_STATUS = True
        self.__evt_start = start_evt or []
        self.__evt_stop = stop_evt or []
        self.__srv = Sanic(self.__conf.SERVER_NAME, log_config=self.__conf.log_conf())
        self.__srv.update_config(self.__conf)
        self.__bp_list = []
        if self.__conf.db_conf():
            from tortoise.contrib.sanic import register_tortoise
            register_tortoise(self.__srv, config=self.__conf.db_conf())
        for mdw_fun in middlewares or []:
            mdw_fun.set_conf(self.__conf)
            self.__srv.middleware(mdw_fun.main)
        self.ws_offline_listen = []
        for bp_cls in bp_arr:
            bp_cls.init(self.__conf.SERVER_NAME, version=self.__conf.VER_CODE)
            bp_cls.load_default_api()
            self.__srv.blueprint(bp_cls.bpo)
            self.ws_offline_listen.extend(bp_cls.ws_handlers)
            bp_cls.ws_handlers.clear()
        for expt_fun in exceptions or []:
            expt_fun.set_conf(self.__conf)
            self.__srv.error_handler.add(Exception, expt_fun.catch_req)
        self.__srv.register_listener(self.__after_server_start, 'after_server_start')
        self.__srv.register_listener(self.__before_server_stop, 'before_server_stop')

    async def __after_server_start(self, app, loop):
        if self.__conf.CONF_RDS:
            for k, cnf in self.__conf.CONF_RDS.items():
                RdsClient.init(cnf, name=k)
        listen_set = set()
        for ws_hd in self.ws_offline_listen:
            loop.create_task(ws_hd.init_offline_queue())
            if hasattr(ws_hd.ws_manager, 'init_loop') and (ws_hd.ws_manager.__name__ not in listen_set):
                ws_hd.ws_manager.set_conf(f"{self.__conf.SERVER_NAME}_{self.__conf.SERVER_ID}", self.proc_name)
                ws_hd.ws_manager.init_loop(rds=RdsClient.clt(), loop=loop)
                listen_set.add(ws_hd.ws_manager.__name__)
        listen_set.clear()
        for fun in self.__evt_start:
            (await fun(**{'app': app, 'loop': loop})) if asyncio.iscoroutinefunction(fun) else fun(**{'app': app, 'loop': loop})
        consts.GLOBAL_SRV_STATUS = True

    async def __before_server_stop(self, app, loop):
        consts.GLOBAL_SRV_STATUS = False
        for fun in self.__evt_stop:
            (await fun(**{'app': app, 'loop': loop})) if asyncio.iscoroutinefunction(fun) else fun(**{'app': app, 'loop': loop})


    @property
    def main(self):
        return self.__srv

    def add_signal(self, signal_map: dict):
        """信号注册"""
        if not signal_map:
            return
        for k, v in signal_map.items():
            callable(v) and self.__srv.add_signal(v, k)

    def add_start_event(self, events):
        if callable(events):
            self.__evt_start.append(events)
        if isinstance(events, (list, tuple)):
            self.__evt_start.extend(events)

    def add_stop_event(self, events):
        if callable(events):
            self.__evt_stop.append(events)
        if isinstance(events, (list, tuple)):
            self.__evt_stop.extend(events)

    def run(self,
            protocol: Optional[type[Protocol]] = None,
            auto_reload: bool=None,
            version: HTTPVersion = HTTP.VERSION_1,
            ssl: Union[None, SSLContext, dict, str, list, tuple] = None,
            sock: Optional[socket] = None,
            backlog: int = 100,
            register_sys_signals: bool = True,
            unix: Optional[str] = None,
            loop: Optional[AbstractEventLoop] = None,
            reload_dir: Optional[Union[list[str], str]] = None,
            noisy_exceptions: Optional[bool] = None,
            motd: bool = True,
            verbosity: int = 0,
            motd_display: Optional[dict[str, str]] = None,
            auto_tls: bool = False,
            single_process: bool = False,
            ):
        """
        Args:

            auto_reload (Optional[bool]): Reload app whenever its source code is changed.
                Enabled by default in debug mode.
            version (HTTPVersion): HTTP Version.
            ssl (Union[None, SSLContext, dict, str, list, tuple]): SSLContext, or location of certificate and key
                for SSL encryption of worker(s).
            sock (Optional[socket]): Socket for the server to accept connections from.
            protocol (Optional[Type[Protocol]]): Subclass of asyncio Protocol class.
            backlog (int): A number of unaccepted connections that the system will allow
                before refusing new connections.
            register_sys_signals (bool): Register SIG* events.
            unix (Optional[str]): Unix socket to listen on instead of TCP port.
            loop (Optional[AbstractEventLoop]): AsyncIO event loop.
            reload_dir (Optional[Union[List[str], str]]): Directory to watch for code changes, if auto_reload is True.
            noisy_exceptions (Optional[bool]): Log exceptions that are normally considered to be quiet/silent.
            motd (bool): Display Message of the Day.
            verbosity (int): Verbosity level.
            motd_display (Optional[Dict[str, str]]): Customize Message of the Day display.
            auto_tls (bool): Enable automatic TLS certificate handling.
            single_process (bool): Enable single process mode.

        Returns:
            None

        Raises:
            RuntimeError: Raised when attempting to serve HTTP/3 as a secondary server.
            RuntimeError: Raised when attempting to use both `fast` and `workers`.
            RuntimeError: Raised when attempting to use `single_process` with `fast`, `workers`, or `auto_reload`.
            TypeError: Raised when attempting to use `loop` with `create_server`.
            ValueError: Raised when `PROXIES_COUNT` is negative.
        """
        self.__srv.run(
            host=self.__conf.HOST,
            port=self.__conf.RUN_PORT,
            workers=self.__conf.RUN_WORKER if not self.__conf.RUN_FAST else 1,
            fast=self.__conf.RUN_FAST,
            debug=self.__conf.DEBUG_MODE,
            dev=self.__conf.DEBUG_MODE,
            access_log=self.__conf.ACCESS_LOG,
            protocol=protocol,
            auto_reload=auto_reload,
            version=version,
            ssl=ssl,
            sock=sock,
            backlog=backlog,
            register_sys_signals=register_sys_signals,
            unix=unix,
            loop=loop,
            reload_dir=reload_dir,
            noisy_exceptions=noisy_exceptions,
            motd=motd,
            motd_display=motd_display,
            verbosity=verbosity,
            auto_tls=auto_tls,
            single_process=single_process,
        )



import datetime
import time
import zoneinfo
from typing import Union

from nsanic.libs import consts

MONTH_MAP = {1: 31, 2: 28, 3: 31, 4: 30, 5: 31, 6: 30, 7: 31, 8: 31, 9: 30, 10: 31, 11: 30, 12: 31}


def check_tz(tz: Union[str, zoneinfo.ZoneInfo] = None):
    """时区处理"""
    if tz and isinstance(tz, str):
        return zoneinfo.ZoneInfo(tz)
    return tz if tz else zoneinfo.ZoneInfo(consts.GLOBAL_TZ)


def check_leap(year: int):
    """检查年份是否是闰年"""
    return (year % 4) if (not year % 100) else (year % 400)


def month_days(year: int, month: int):
    """获取月份里的所有天数"""
    if check_leap(year) and month == 2:
        return 29
    return MONTH_MAP.get(month)


def cur_dt(tz: Union[str, zoneinfo.ZoneInfo] = None):
    """当前日期"""
    return datetime.datetime.now(tz=check_tz(tz))


def cur_time(ms=False):
    """当前时间戳(utc)"""
    return int(time.time() * 1000) if ms else int(time.time())


def cut_utctime(ms=False):
    """当前时间戳 UTC"""
    return int(time.time()) if not ms else int(time.time() * 1000)


def dt_str(dt: Union[datetime.datetime, datetime.date, int, float, str] = None,
           fmt='%Y-%m-%d %H:%M:%S',
           tz: Union[str, zoneinfo.ZoneInfo] = None):
    """日期时间字符串格式输出, 不指定时间将输出当前时间"""
    if not dt:
        return cur_dt(tz=tz).strftime(fmt)
    if isinstance(dt, datetime.datetime):
        return dt.strftime(fmt)
    elif isinstance(dt, datetime.date):
        fmt = (fmt.replace('%H:', '').replace('%H', '').replace('%M:', '').
               replace('%M', '')).replace('%S.', '').replace('%S', '').replace('%f', '')
        return dt.strftime(fmt)
    elif isinstance(dt, (int, float)):
        dt = datetime.datetime.fromtimestamp(dt, tz=check_tz(tz))
        return dt.strftime(fmt)
    elif isinstance(dt, str):
        dt.replace('/', '-').replace('T', ' ')
        if len(dt) > 10:
            return datetime.datetime.strptime(dt, '%Y-%m-%d %H:%M:%S').strftime(fmt)
        return datetime.datetime.strptime(dt, '%Y-%m-%d').strftime(fmt)


def to_datetime(dt: Union[datetime.date, datetime.datetime, str, int, float],
                fmt='%Y-%m-%d %H:%M:%S', tz: Union[str, zoneinfo.ZoneInfo] = None):
    """将字符、数字等格式转化为时间日期格式"""
    if dt is None:
        return
    if isinstance(dt, str):
        dt = datetime.datetime.strptime(dt, fmt)
    if isinstance(dt, datetime.datetime):
        return dt
    elif isinstance(dt, datetime.date):
        return datetime.datetime.combine(dt, datetime.time(), tzinfo=check_tz(tz))
    elif isinstance(dt, (int, float)):
        return datetime.datetime.fromtimestamp(dt, tz=check_tz(tz))
    return


def create_dt(
        year=2023,
        month=1,
        day=1,
        hour=0,
        minute=0,
        second=0,
        misecond=0,
        tz: Union[str, zoneinfo.ZoneInfo] = None):
    """指定创建时间日期"""
    return datetime.datetime(
        year=year, month=month, day=day, hour=hour, minute=minute, second=second,
        microsecond=misecond, tzinfo=check_tz(tz))


def day_begin(dt: Union[datetime.datetime, datetime.date] = None, tz: Union[str, zoneinfo.ZoneInfo] = None):
    """指定时间/当天的开始时间戳 """
    if not dt:
        dt = cur_dt(tz=tz)
    if isinstance(dt, datetime.datetime):
        dt = dt.replace(hour=0, minute=0, second=0, microsecond=0)
        return int(dt.timestamp())
    return int(datetime.datetime.combine(dt, datetime.time(), tzinfo=check_tz(tz)).timestamp())


def day_end(dt: Union[datetime.datetime, datetime.date] = None, tz: Union[str, zoneinfo.ZoneInfo] = None):
    """指定时间/当天的结束时间戳 """
    if not dt:
        dt = cur_dt(tz=tz)
    if isinstance(dt, datetime.date):
        dt = datetime.datetime.combine(dt, datetime.time(), tzinfo=check_tz(tz))
    dt = dt.replace(hour=23, minute=59, second=59, microsecond=999999)
    return int(dt.timestamp())


def month_begin(dt: Union[datetime.datetime, datetime.date] = None, tz: Union[str, zoneinfo.ZoneInfo] = None):
    """指定时间/当月的开始时间戳"""
    if not dt:
        dt = cur_dt(tz=tz)
    if isinstance(dt, datetime.datetime):
        dt = dt.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        return int(dt.timestamp())
    ndt = datetime.datetime.combine(dt, datetime.time(), tzinfo=check_tz(tz))
    ndt.replace(day=1)
    return int(ndt.timestamp())


def day_interval(dt: Union[datetime.datetime, datetime.date, int, float] = None,
                 tz: Union[str, zoneinfo.ZoneInfo] = None):
    """指定时间/当天的开始与结束的时间戳"""
    if not dt:
        dt = cur_dt(tz=tz)
    if isinstance(dt, (int, float)):
        dt = datetime.datetime.fromtimestamp(dt, tz=check_tz(tz))
    if isinstance(dt, datetime.date):
        dt = datetime.datetime.combine(dt, datetime.time(), tzinfo=check_tz(tz))
    s_dt = dt.replace(hour=0, minute=0, second=0, microsecond=0)
    e_dt = dt.replace(hour=23, minute=59, second=59, microsecond=999999)
    return int(s_dt.timestamp()), int(e_dt.timestamp())


def day_hours(dt: Union[datetime.datetime, datetime.date, int, float] = None, tz: Union[str, zoneinfo.ZoneInfo] = None):
    """一天开始到结束按小时拆分的时间戳"""
    st, et = day_interval(dt, tz=tz)
    hour_list = []
    start = st
    for i in range(st, et, 3600):
        end = start + 3600 - 1
        if end >= et:
            end = et
        hour_list.append((start, end))
        start = end + 1
    return hour_list


def date_range(
        start: Union[datetime.datetime, datetime.date] = None,
        end: Union[datetime.datetime, datetime.date] = None,
        days: int = 0,
        tz: Union[str, zoneinfo.ZoneInfo] = None):
    """输出指定日期范围"""
    if isinstance(start, datetime.datetime):
        start = start.date()
    if isinstance(end, datetime.datetime):
        end = end.date()
    if start and end:
        if start > end:
            start, end = end, start
        delta = end - start
        return [start + datetime.timedelta(days=i) for i in range(delta.days + 1)]
    if start and days:
        return [start + datetime.timedelta(days=i) for i in range(days + 1)]
    if end and days:
        return [end - datetime.timedelta(days=i) for i in range(days + 1)]
    return [start] if start else ([end] if end else [cur_dt(tz=tz).date()])


def day_before(days: int, set_date: Union[datetime.datetime, datetime.date] = None, tz: Union[str, zoneinfo.ZoneInfo] = None):
    if not set_date:
        set_date = cur_dt(tz=tz)
    return (set_date - datetime.timedelta(days)) if days > 0 else (set_date + datetime.timedelta(abs(days)))



import asyncio
import traceback
from typing import Union

from redis.asyncio import Redis, ConnectionPool, BlockingConnectionPool

from nsanic.libs.component import LogMeta
from nsanic.libs.rds_locker import RdsLocker
from nsanic.libs.tool import json_encode, json_parse


class RdsError(Exception):
    pass


class RdsClient(LogMeta):
    """redis连接对象 单例模式请用init方法初始化"""
    CONN_MAP = {}

    def __init__(self, conf: dict):
        self.__timeout = conf.pop('timeout') if 'timeout' in conf else 10
        self.__ping = conf.pop('ping_interval') if 'ping_interval' in conf else 60
        self.__block = conf.pop('use_block') if 'use_block' in conf else False
        self.is_decode = conf.get('decode_responses', False)
        self.__conf = conf
        self.__pool = ConnectionPool(**self.__conf) if not self.__block else BlockingConnectionPool(
            **self.__conf, timeout=self.__timeout)

    @classmethod
    def clt(cls, name='default'):
        if name not in cls.CONN_MAP:
            raise Exception(f"当前连接配置未初始化: {name}")
        return cls.CONN_MAP[name]

    @classmethod
    def init(cls, conf: dict, name: Union[str, int, float, bytes] = None):
        """单例模型请用该方法进行初始化(请在事件循环里调用该方法)"""
        if not all(key in conf for key in ['host', 'port', 'db']):
            raise Exception('Redis连接配置错误,缺少必要的连接配置项')
        clt = cls.CONN_MAP.get(name)
        if not clt:
            clt = cls(conf)
            clt.init_loop()
            cls.CONN_MAP[name] = clt
        return clt

    def init_loop(self, loop=None):
        if not loop:
            loop = asyncio.get_event_loop()
        loop.create_task(self.__loop_ping())

    async def __loop_ping(self):
        while 1:
            try:
                async with Redis(connection_pool=self.__pool) as clt:
                    await clt.ping()
            except Exception as err:
                self.logerr(f'redis连接已失效,将在{self.__timeout}秒后重试: {err} \n{traceback.format_exc()}')
                self.__pool.reset()
                await asyncio.sleep(self.__timeout)
                continue
            await asyncio.sleep(self.__ping)

    @property
    def conn(self):
        """使用该方式请调用close()手动释放连接"""
        return Redis(connection_pool=self.__pool)

    def reset_pool(self):
        """重置连接池 慎用"""
        self.__pool and self.__pool.reset()

    async def locked(self, key: str, fun=None, fun_param: Union[list, tuple] = None, time_out=5, pre_key='LOCKER__'):
        """
        :param key: 锁定标识
        :param fun: 执行函数
        :param fun_param: 函数参数
        :param time_out: 自动释放时间
        :param pre_key: 锁定标识预制前缀，用于区分锁定标识采用redis key本身
        """
        if not fun_param:
            fun_param = ()
        if pre_key:
            key = f'{pre_key}{key}'
        async with Redis(connection_pool=self.__pool) as clt:
            async with clt.lock(key, lock_class=RdsLocker, timeout=time_out):
                return callable(fun) and ((await fun(*fun_param)) if asyncio.iscoroutinefunction(fun) else fun(*fun_param))

    async def get_item(self, key: Union[str, bytes], jsparse=False):
        """
        获取缓存
        :param key: 缓存名称
        :param jsparse: 是否解析Json字符串为字典
        """
        async with Redis(connection_pool=self.__pool) as clt:
            info = await clt.get(key)
            if jsparse:
                return json_parse(info)
            return info.encode() if self.is_decode else info

    async def del_item(self, *args):
        """删除"""
        if not args:
            return
        async with Redis(connection_pool=self.__pool) as clt:
            return await clt.delete(*args)

    async def set_item(self, key: Union[str, bytes], value: Union[str, int, float, bytes, list, tuple, dict],
                       ex_time: int = None):
        """
        设置字符串缓存
        :param key: 缓存名称
        :param value: 缓存值，非字符串类型将会自动转换为字符串存储
        :param ex_time: 缓存有效期，默认不设置，既永久有效，类型数字类型 单位：秒 # type:int
        """
        if isinstance(value, (list, tuple, dict)):
            value = json_encode(value)
        async with Redis(connection_pool=self.__pool) as clt:
            return await clt.set(key, value, ex=ex_time)

    async def exists(self, key: Union[str, bytes]):
        """检查key是否存在"""
        async with Redis(connection_pool=self.__pool) as clt:
            return await clt.exists(key)

    async def hexists(self, name: Union[str, bytes], key):
        """检查key是否存在"""
        async with Redis(connection_pool=self.__pool) as clt:
            return await clt.hexists(name, key)

    async def drop_item(self, key: Union[str, bytes]):
        """
        根据key删除某项缓存
        :param key: 缓存名称
        """
        async with Redis(connection_pool=self.__pool) as clt:
            return await clt.delete(key)

    async def get_hash(self, key: Union[str, bytes], h_key: Union[str, bytes], jsparse=False):
        """
        获取Hash值

        :param key: 缓存名称
        :param h_key: 缓存键名称
        :param jsparse: 是否解析Json字符串为字典
        """
        async with Redis(connection_pool=self.__pool) as clt:
            info: Union[str, bytes] = await clt.hget(key, h_key)
            if jsparse:
                return json_parse(info)
            return info.encode() if self.is_decode else info

    async def set_hash(self, key: Union[str, bytes], h_key: Union[str, bytes], data):
        """
        设置Hash

        :param key: 缓存名称
        :param h_key: 缓存键名称
        :param data: 保存的数据
        """
        if isinstance(data, list) or isinstance(data, dict):
            data = json_encode(data, log_fun=self.logerr)
        async with Redis(connection_pool=self.__pool) as clt:
            return await clt.hset(key, h_key, data)

    async def drop_hash(self, key: Union[str, bytes], h_key: Union[str, bytes]):
        """
        删除Hash值
        :param key: 缓存名称
        :param h_key: 缓存键名称
        """
        async with Redis(connection_pool=self.__pool) as clt:
            return await clt.hdel(key, h_key)

    async def drop_hash_bulk(self, key: Union[str, bytes], key_list: Union[list, tuple]):
        """
        批量删除Hash值
        :param key: 缓存名称
        :param key_list: 缓存键列表
        """
        async with Redis(connection_pool=self.__pool) as clt:
            key_list and (await clt.hdel(key, *key_list))

    async def get_hash_all(self, key: Union[str, bytes], jsparse=False):
        """
        获取整个Hash
        :param key: 缓存名
        :param jsparse: 是否解析Json字符串为字典
        """
        async with Redis(connection_pool=self.__pool) as clt:
            info = await clt.hgetall(key)
            if not info:
                return
            if jsparse:
                new_dict = {}
                for k, v in info.items():
                    if isinstance(k, bytes):
                        k = k.decode()
                    if jsparse:
                        v = json_parse(v)
                    new_dict[k] = v
                return new_dict
            return info

    async def get_hash_val(self, key: Union[str, bytes]):
        """
        获取整个Hash所有值
        :param key: 缓存名
        """
        async with Redis(connection_pool=self.__pool) as clt:
            return await clt.hvals(key)

    async def set_hash_bulk(self, key: Union[str, bytes], h_map: dict):
        """
        通过字典映射设置或更新多个Hash
        :param key: 缓存名
        :param h_map: 字典映射 -- 必须是{str: str} 或{bytes: str} 或{bytes: bytes} 几种形式
        """
        async with Redis(connection_pool=self.__pool) as clt:
            return await clt.hset(key, mapping=h_map)

    async def scan_hash(self, key: Union[str, bytes], start: int = 0, count: int = None, match: str = None):
        """
        按匹配的条件迭代扫描hash
        """
        async with Redis(connection_pool=self.__pool) as clt:
            return await clt.hscan(key, cursor=start, match=match, count=count)

    async def pub_sub(self, channel: Union[str, list, tuple]):
        """消息订阅模型"""
        async with Redis(connection_pool=self.__pool) as clt:
            sub_item = clt.pubsub()
            if isinstance(channel, str):
                channel = [channel]
            await sub_item.subscribe(*channel)
            return sub_item

    async def publish(self, channel: str, msg: Union[bytes, str]):
        """
        发布消息
        :param channel: 消息信道
        :param msg: 发布的消息
        """
        async with Redis(connection_pool=self.__pool) as clt:
            return await clt.publish(channel, msg)

    async def qlpush(self, key: str, data_list: list):
        """
        左侧入队列
        :param key: 队列/集合名称
        :param data_list: 入队列表 --type: list
        """
        values = []
        for item in data_list:
            if isinstance(item, dict) or isinstance(item, list):
                item = json_encode(item, log_fun=self.logerr)
            values.append(item)
        async with Redis(connection_pool=self.__pool) as clt:
            return await clt.lpush(key, *values)

    async def qrpush(self, key: str, data_list: list):
        """
        右侧入队列
        :param key: 队列/集合名称
        :param data_list: 入队列表 --type: list
        """
        values = []
        for item in data_list:
            if isinstance(item, dict) or isinstance(item, list):
                item = json_encode(item, log_fun=self.logerr)
            values.append(item)
        async with Redis(connection_pool=self.__pool) as clt:
            return await clt.rpush(key, *values)

    async def qrpop(self, key: str, count=1, jsparse=False):
        """
        右侧出队

        :param key: 队列/集合名称
        :param count: 出队数量
        :param jsparse: 是否解析Json字符串为字典
        """
        async with Redis(connection_pool=self.__pool) as clt:
            info: Union[str, bytes, list] = await clt.rpop(key, count=None if count <= 1 else count)
            if not info:
                return
            if count <= 1:
                if jsparse:
                    return json_parse(info)
                return info.encode() if self.is_decode else info
            if jsparse:
                return [json_parse(item) for item in info]
            return [item.encode() for item in info] if self.is_decode else info

    async def qlpop(self, key: str, count=1, jsparse=False):
        """
        左侧出队

        :param key: 队列/集合名称
        :param count: 出队数量
        :param jsparse: 是否解析Json字符串为字典
        """
        async with Redis(connection_pool=self.__pool) as clt:
            info: Union[str, bytes, list] = await clt.lpop(key, count=None if count <= 1 else count)
            if not info:
                return
            if count <= 1:
                if jsparse:
                    return json_parse(info)
                return info.encode() if self.is_decode else info
            if jsparse:
                return [json_parse(item) for item in info]
            return [item.encode() for item in info] if self.is_decode else info

    async def qlen(self, key: str):
        """
        获取队列长度
        :param key: 队列/集合名称
        """
        async with Redis(connection_pool=self.__pool) as clt:
            return await clt.llen(key)

    async def expired(self, key: Union[str, bytes], t: int, nx=False, xx=False, gt=False, lt=False):
        async with Redis(connection_pool=self.__pool) as clt:
            return await clt.expire(key, time=t, nx=nx, xx=xx, gt=gt, lt=lt)

    async def incr(self, key: Union[str, bytes], count: int):
        async with Redis(connection_pool=self.__pool) as clt:
            return await clt.incr(key, int(count))

    async def incrby(self, key: Union[str, bytes], count: int = 1):
        async with Redis(connection_pool=self.__pool) as clt:
            return await clt.incrby(key, int(count))

    async def pipline_fun(self, fun, transaction=True, raise_err=True):
        """
        :param fun: 指定添加pipline的方法，唯一参数为pipline对象
        :param transaction: 是否使用事务执行，redis事务不支持回滚
        :param raise_err: 有失败是否报错
        """
        async with Redis(connection_pool=self.__pool) as clt:
            pl = clt.pipeline(transaction=transaction)
            fun(pl)
            if len(pl):
                await pl.execute(raise_on_error=raise_err)
                await clt.close()



import asyncio

from redis.asyncio.lock import Lock


class RdsLocker(Lock):

    _watcher: asyncio.Task

    async def __waiting(self):
        while 1:
            await self.extend(self.timeout)
            await asyncio.sleep(self.timeout/3)

    def __stop_watcher(self):
        old: asyncio.Future = getattr(self, "_watcher", None)
        if old and (not old.cancelled()):
            old.cancel()

    async def acquire(self, blocking: bool = None, timeout: float = None, token: (str, bytes) = None) -> bool:
        result = await super().acquire(blocking, timeout, token)
        if result:
            self.__stop_watcher()
            self._watcher = asyncio.create_task(self.__waiting())
        return result

    async def do_release(self, expected_token: bytes) -> None:
        self.__stop_watcher()
        return await super().do_release(expected_token)

    def __del__(self) -> None:
        try:
            self.__stop_watcher()
        except Exception as err:
            _ = err
            pass



from typing import Union

from nsanic.exception import WsRpsMsg
from nsanic import verify
from nsanic.libs.component import LogMeta
from nsanic.libs.consts import Code, StaCode


class WsHandler(LogMeta):
    STA_CODE = StaCode

    CMD_FUN_MAP = {}

    def __init__(self, cmd_type: Union[int, str, float]):
        self.__cmd_type = cmd_type

    @property
    def cmd_type(self):
        return self.__cmd_type

    async def funapi(self, fcmd, uinfo, data):
        fun = self.CMD_FUN_MAP.get(fcmd)
        if fun:
            try:
                await fun(uinfo, data)
            except WsRpsMsg as msg:
                return [msg.end, msg.mult, fcmd, msg.data, msg.code, msg.hint]
            else:
                return [False, False, fcmd, data, self.STA_CODE.FAIL, '']
        return [False, False, fcmd, None, self.STA_CODE.FAIL, 'unspecified message.']

    async def offline(self, ukey: Union[int, str]):
        """掉线处理逻辑"""
        pass

    def sendmsg(self, code: Code = None, data: Union[dict, list] = None, mult=False, end=False, hint=''):
        """
        消息下发
        :param code: 状态码
        :param data: 消息数据，同类多消息结构可用list 不同类消息请使用{cmd:message}映射关系
        :param mult: 是否为多重消息
        :param end: 是否终止连接
        :param hint: 提示信息转发
        """
        if not code:
            code = self.STA_CODE.PASS
        raise WsRpsMsg(code, data, mult=mult, end=end, hint=hint)

    def vf_str(self, val, require=False, default='', turn=0, minlen: int = None, maxlen: int = None, p_name=''):
        sta, val_info = verify.vstr(
            val, require=require, default=default, turn=turn, minlen=minlen, maxlen=maxlen, p_name=p_name)
        return val_info if sta else self.sendmsg(code=self.STA_CODE.ERR_ARG, hint=val_info)

    def vf_int(self, v, require=False, default=0, minval: int = None, maxval: int = None, inner=True, p_name=''):
        sta, val_info = verify.vint(
            v, require=require, default=default, minval=minval, maxval=maxval, inner=inner, p_name=p_name)
        return val_info if sta else self.sendmsg(code=self.STA_CODE.ERR_ARG, hint=val_info)

    def vf_float(self, val, require=False, default=None, keep_val=3, minval: float or int = None,
                 maxval: float or int = None, inner=True, p_name=''):
        sta, val_info = verify.vfloat(
            val, require=require, default=default, keep_val=keep_val, minval=minval, maxval=maxval,
            inner=inner, p_name=p_name)
        return val_info if sta else self.sendmsg(code=self.STA_CODE.ERR_ARG, hint=val_info)

    def vf_time(self, val, require=False, default=None, time_min=None, time_max=None, inner=True, p_name=''):
        sta, val_info = verify.vdatetime(
            val, require=require, default=default, time_min=time_min, time_max=time_max, inner=inner, p_name=p_name)
        return val_info if sta else self.sendmsg(code=self.STA_CODE.ERR_ARG, hint=val_info)



import traceback
from dataclasses import dataclass
from typing import Union

from sanic import Request, response

from nsanic.libs.consts import Code
from nsanic.libs.manager import HeaderSet
from nsanic.libs.mult_log import NLogger
from nsanic.libs.rds_client import RdsError
from nsanic.libs.tool import json_encode


@dataclass
class JsonFinish(Exception):
    """完成请求状态异常"""
    code: Code
    '''状态码'''
    data: Union[dict, list, str, int, float] = None
    '''响应数据'''
    total: int = 0
    '''针对于分页数据响应的总数量'''
    hint: str = ''
    '''响应消息提示'''
    headers: dict = None
    '''附加响应头'''


@dataclass
class WsRpsMsg(Exception):
    """websocket消息响应"""
    code: Code
    '''状态码'''
    data: Union[dict, list, str, int, float] = None
    '''响应数据'''
    hint: str = ''
    '''响应消息提示'''
    mult: bool = False
    '''是否为多重消息模式'''
    end: bool = False
    '''是否终止连接'''


class CatchExpt:

    code_map = {
        400: '请求结构错误,无法解析的结构',
        404: 'API接口不存在',
        500: '出错啦，请稍后再试或联系管理员检查',
        403: '请求结构不支持'
    }

    @classmethod
    def set_conf(cls, conf):
        cls.conf = conf

    @classmethod
    def __loger(cls, req: Request, err_info: str):
        NLogger.error(
            f"请求异常信息:\n接口--{req.path}\n方法--{req.method}\n头部--{req.headers.items()}\nquery参数--{req.args}\n"
            f"body--{str(req.body, 'utf8')}\n错误信息--{err_info}")

    @classmethod
    def catch_req(cls, req: Request, expt):
        headers = HeaderSet.out(cls.conf)
        hasattr(expt, 'headers') and expt.headers and headers.update(expt.headers)
        if isinstance(expt, RdsError):
            body = {'code': 500, 'data': None, 'msg': 'Invalid Cache Server. Please contact administrator.'}
            return response.json(body, status=500, headers=headers, dumps=json_encode)
        if isinstance(expt, JsonFinish):
            body = {'code': expt.code.val, 'data': expt.data, 'msg': expt.hint or expt.code.msg}
            return response.json(body, status=expt.code.http, headers=headers)
        body = {'code': 500, 'data': None, 'msg': 'There are some error, please connect administrator to check.'}
        if hasattr(expt, 'status_code'):
            hint = cls.code_map.get(expt.status_code)
            body = {'code': expt.status_code, 'data': None, 'msg': hint}
            if expt.status_code == 500:
                cls.__loger(req, traceback.format_exc())
            return response.json(body, headers=headers)
        cls.__loger(req, traceback.format_exc())
        return response.json(body, status=500)



from tortoise.indexes import Index


class VchpIndex(Index):
    """BTree类索引 模糊匹配仅支持xxx% varchar_pattern模式通配符 简单通配下配合limit offset使用达到最好性能"""
    INDEX_CREATE_TEMPLATE = (
        "CREATE{index_type}INDEX {index_name} ON {table_name} ({fields} varchar_pattern_ops){extra};")


class GistTrgmIndex(Index):
    """
    gist模糊匹配索引
    支持模糊匹配 %%全通配符 实测简单varchar char模式下十万百万级别的数据下性能更差 另外需要安装pg_trgm插件
    """

    INDEX_TYPE = 'GIST'
    INDEX_CREATE_TEMPLATE = (
        "CREATE INDEX {index_name} ON {table_name} USING {index_type}({fields} gist_trgm_ops){extra};")



from datetime import datetime, date
from typing import Union

from nsanic.libs.component import LogMeta
from nsanic.libs.consts import GLOBAL_TZ
from nsanic.libs.rds_client import RdsClient
from nsanic.libs.tool import json_parse, json_encode
from nsanic.orm.db_model import DBModel


class RCModel(LogMeta):
    """数据缓存redis模型"""
    rds: RdsClient = None
    db_model: DBModel = None
    '''数据模型'''
    expired_mode = 0
    '''缓存过期模式 0--hash类的固定缓存(无过期) 1--key value类的时效型缓存 带过期时间 默认12小时'''
    expired_sec = 43200
    '''key value类的时效型缓存的过期时间'''

    @classmethod
    async def fun_set_cache(cls, info: dict):
        """附加缓存设置"""
        pass

    @classmethod
    async def cache_by_pk(cls, pk_val: Union[bytes, int, str], split: Union[str, int, float, datetime, date] = None,
                          tz: str = GLOBAL_TZ):
        async def from_db():
            db_info = await cls.db_model.get_by_pk(pk_val, tz=tz)
            if db_info:
                if cls.expired_mode:
                    await cls.rds.set_item(key, json_encode(db_info), ex_time=cls.expired_sec)
                else:
                    await cls.rds.set_hash(tb_name, pk_val, json_encode(db_info))
                await cls.fun_set_cache(db_info)
                return db_info
            return

        if not cls.db_model:
            return
        tb_name = cls.db_model.sheet_name(split, tz=tz)
        if isinstance(pk_val, bytes):
            pk_val = pk_val.decode('utf-8')
        key = f"{tb_name}:{pk_val}"
        if cls.expired_mode:
            info = await cls.rds.get_item(key)
        else:
            info = await cls.rds.get_hash(tb_name, pk_val)
        if info:
            return json_parse(info, cls.logerr)
        p_info = await cls.rds.locked(key, from_db)
        return p_info

    @classmethod
    async def cache_by_unique(cls, unique: dict, suffix: str, split: Union[str, int, float, datetime, date] = None,
                              tz: str = GLOBAL_TZ):
        async def from_db():
            info = await cls.db_model.get_by_dict(unique, limit=1, split=split, tz=tz)
            if info:
                pk_val = info.get(cls.db_model.pk_name())
                if cls.expired_mode:
                    await cls.rds.set_item(key, pk_val, ex_time=cls.expired_sec)
                else:
                    await cls.rds.set_hash(key_name, unique_val, pk_val)
                await cls.fun_set_cache(info)
                return info
            return

        if not cls.db_model:
            return
        key_name = f'{cls.db_model.sheet_name(split)}_{suffix}'
        unique_val = '_'.join([str(unique.get(k)) for k in sorted(unique)])
        key = f'{key_name}:{unique_val}'
        if cls.expired_mode:
            item_id = await cls.rds.get_item(key)
        else:
            item_id = await cls.rds.get_hash(key_name, unique_val)
        if item_id:
            return await cls.cache_by_pk(item_id, split=split, tz=tz)
        return await cls.rds.locked(key, fun=from_db)

    # @classmethod
    # async def query_map(cls, is_super=False, field_name='name', groups: (list, tuple) = None, group_key='gid'):
    #     """"""
    #     async def from_db():
    #         data_list = await cls.db_model.get_by_dict()
    #         return [await cls.fun_set_cache(info) for n in data_list]
    #     if not groups:
    #         groups = []
    #     if not cls.db_model:
    #         return []
    #     pk_name = cls.db_model.pk_name()
    #     all_items = await cls.conf.rds.get_hash_val(cls.db_model.sheet_name())
    #     if not all_items:
    #         all_items = await cls.conf.rds.locked(cls.db_model.sheet_name(), fun=from_db, time_out=5)
    #     if (not is_super) and cls.db_model.check_field(group_key):
    #         new_list = []
    #         for item in all_items:
    #             info = json_parse(item, log_fun=cls.logerr)
    #             g_id = info.get(group_key)
    #             (g_id in groups) and new_list.append({
    #                 'label': info.get(field_name), 'value': info.get(pk_name), group_key: g_id})
    #         return new_list
    #     return [{'label': info.get(field_name), 'value': info.get(pk_name), group_key: info.get(group_key)}
    #             for item in all_items if (info := json_parse(item, cls.logerr))]

    @classmethod
    async def get_name(cls, pk_id: int or str, key_field='name', split: Union[str, int, float, datetime, date] = None,
                       tz: str = GLOBAL_TZ):
        if not pk_id:
            return '-'
        info = await cls.cache_by_pk(pk_id, split=split, tz=tz)
        return info.get(key_field) if info else '-'

    @classmethod
    async def get_map(cls, pk_id: int or str, key_list: list or tuple = None,
                      split: Union[str, int, float, datetime, date] = None, tz: str = GLOBAL_TZ):
        if not pk_id:
            return {}
        info = await cls.cache_by_pk(pk_id, split=split, tz=tz)
        return {name: info.get(name) for name in key_list} if info else {}



from nsanic.libs.base_timed import BaseTimed


class TestTImed(BaseTimed):

    def __init__(self):
        super().__init__()
        self.add_loop_task('test01', 10, self.test_01)
        self.add_timed_task('test02', ['*', '*', '*', '*', '*', 30], self.test_02)

    async def test_01(self):
        print('test_01')

    async def test_02(self):
        print('test_02')


a = TestTImed()
a.start()



import datetime
import re

from nsanic.libs.tool_dt import to_datetime, dt_str


def password(pwd_str: str, len_min=6, len_max=18, qc=2):
    """
    统一密码校验
    :param pwd_str: 待校验的字符串
    :param len_min: 最小长度
    :param len_max: 最大长度
    :param qc: 密码复杂度 取值1-4
    """
    if not isinstance(pwd_str, str):
        pwd_str = str(pwd_str)
    if len_max >= len(pwd_str) >= len_min:
        sign = [0, 0, 0, 0]
        for i in pwd_str:
            ord_id = ord(i)
            if 33 <= ord_id <= 126:
                if 48 <= ord_id <= 57:
                    sign[0] = 1
                elif 65 <= ord_id <= 90:
                    sign[1] = 1
                elif 97 <= ord_id <= 122:
                    sign[2] = 1
                else:
                    sign[3] = 1
            else:
                return 0, f'The password contains invalid characters:{i}'
        if sum(sign) < qc:
            return 0, 'The password should contain at least two types of digits, uppercase, lowercase, special symbols'
        return 1, ''
    return 0, 'The password should be 6 to 18 characters long'


def common_str(item_str: str, start_letter=True):
    """常规字符串校验 以字母开头 只包含数字字母下划线"""
    num_letter = re.compile("^[0-9a-zA-Z_]+$")
    if num_letter.search(item_str):
        if start_letter:
            first = re.compile('^[a-zA-Z]')
            if first.search(item_str):
                return True
        else:
            return True
    return False


def color(color_str: str):
    return True if re.match(r'^#[A-F0-9]', color_str) else False


def phone(phone_str: str):
    return True if re.match(r'^1[3-9]\d{9}', phone_str) else False


def extra_model(new_model: dict, default: dict, old_model: dict = None):
    """统一处理扩展字段检查"""
    if (not default) and (not old_model):
        return new_model
    if not old_model:
        old_model = {}
    if not default:
        default = {}
    new_extra = {}
    for k in new_model:
        v = default.get(k)
        old_v = old_model.get(k)
        new_extra[k] = v if old_v != v else old_v
    return new_extra


def vint(val, require=False, default=None, minval: int = None, maxval: int = None, inner=True, p_name=''):
    """
    整形参数校验

    :param val: 待校验值
    :param require: 是否必要参数 默认非必要
    :param default: 非必要状态下的默认值
    :param minval: 最小值范围 type--int 默认不校验
    :param maxval: 最大值范围 type--int 默认不校验
    :param inner: 是否范围内(针对于min_val或max_val有值校验)，默认范围内，False为范围外
    :param p_name: 参数名
    :return 状态, 值/失败原因
    """
    if val in ('', None, 'None', [], {}, 'null', 'NULL', 'Null'):
        if require:
            return False, f'Miss required parameter：{p_name}'
        return True, default
    if not isinstance(val, int):
        try:
            if isinstance(val, str):
                val = val.split('.')[0]
            val = int(val)
        except(ValueError, TypeError, SyntaxError):
            return False, f"The parameter '{p_name}' cannot be resolved as an integer"
    if inner:
        if (minval is not None) and (val < minval):
            return False, f"The parameter '{p_name}' is less than the minimum value of the limit: {minval}"
        if (maxval is not None) and (val > maxval):
            return False, f"The parameter '{p_name}' is greater than the maximum value: {maxval}"
    else:
        if (minval is not None) and (maxval is not None) and (minval < val < maxval):
            return False, f"The parameter '{p_name}' is outside the specified limit"
    return True, val


def vstr(val, require=False, default: str = None, turn=0, minlen: int = None, maxlen: int = None, p_name=''):
    """
    字符串参数校验

    :param val: 待校验对象
    :param require: 是否必要参数 默认非必要
    :param default: 非必要状态下的默认值
    :param turn: 默认转换 1--全转化大写 2--全转换小写 3--首字母大写 其它值--不转化
    :param minlen: 最小长度 type--int 默认不校验
    :param maxlen: 最大长度 type--int 默认不校验
    :param p_name: 参数名
    """
    if isinstance(val, str):
        val = val.strip()
    if val in ('', None, 'None', [], {}, 'null', 'NULL', 'Null'):
        if require:
            return False, f'Miss required parameter：{p_name}'
        return True, default
    if not isinstance(val, str):
        val = str(val)
    len_val = len(val)
    if turn == 1:
        val = val.upper()
    elif turn == 2:
        val = val.lower()
    elif turn == 3:
        val = f'{val[0].upper()}{val[1:]}' if len_val > 1 else val.upper()
    if (not minlen) and (not maxlen):
        return True, val
    if minlen and (len_val < minlen):
        return False, f"Parameter '{p_name}' length is less than the minimum limit"
    if maxlen and (len_val > maxlen):
        return False, f"Parameter '{p_name}' length is greater than the limit"
    return True, val


def vfloat(
        val,
        require=False,
        default: float = None,
        keep_val=3,
        minval: float or int = None,
        maxval: float or int = None,
        inner=True,
        p_name=''):
    """
    浮点数校验

    :param val: 待校验对象
    :param require: 是否必要参数 默认非必要
    :param default: 非必要状态下的默认值
    :param keep_val: 保留小数未 默认3位
    :param minval: 最小值范围 type--int 默认不校验
    :param maxval: 最大值范围 type--int 默认不校验
    :param inner: 是否范围内(针对于min_val或max_val有值校验)，默认范围内，False为范围外
    :param p_name: 参数名
    :return 校验状态--bool, 转换的值--float/错误信息
    """
    if val in ('', None, 'None', [], {}, 'null', 'NULL', 'Null'):
        if require:
            return False, f'Miss required parameter：{p_name}'
        return True, default
    try:
        val = float('%.{0}f'.format(keep_val) % float(val))
    except(SyntaxError, ValueError, TypeError):
        return False, f"The parameter '{p_name}' cannot be resolved as a float"
    if (minval is None) and (maxval is None):
        return True, val
    if inner:
        if (maxval is not None) and (val > maxval):
            return False, f"The parameter '{p_name}' is greater than the maximum value: {maxval}"
        if (minval is not None) and (val < minval):
            return False, f"The parameter '{p_name}' is less than the minimum value of the limit: {minval}"
    else:
        if (minval is not None) and (maxval is not None) and (minval < val < maxval):
            return False, f"The parameter '{p_name}' is outside the specified limit"
    return True, val


def vdatetime(
        val,
        require=False,
        default: datetime.datetime = None,
        time_min: datetime.datetime = None,
        time_max: datetime.datetime = None,
        inner=True,
        p_name=''):
    """
    校验时间

    :param val: 待校验对象
    :param require: 是否必要参数 默认非必要
    :param default: 非必要状态下的默认值
    :param time_min: 最小值范围 type--Datetime 默认不校验
    :param time_max: 最大值范围 type--Datetime 默认不校验
    :param inner: 是否范围内(针对于min_val或max_val有值校验)，默认范围内，False为范围外
    :param p_name: 参数名
    :return 校验状态--bool, 转换的值--float/错误信息
    """
    if val in ('', None, 'None', [], {}, 'null', 'NULL', 'Null'):
        if require:
            return False, f'Miss required parameter：{p_name}'
        return True, default
    val_time = to_datetime(val)
    if not val_time:
        return False, f"The parameter '{p_name}' cannot be resolved as a time type"
    if (not time_min) and (not time_max):
        return True, val_time
    if inner:
        if time_min and val_time < time_min:
            return False, f"The parameter '{p_name}' is outer of the minimum time: {dt_str(time_min)}"
        if time_max and val_time > time_max:
            return False, f"The parameter '{p_name}' is outer of the maximum time: {dt_str(time_max)}"
    else:
        if time_min and time_max and (time_min < val_time < time_max):
            return False, f"The parameter '{p_name}' is outer of the time range"
    return True, val



import time
import traceback
from functools import wraps
from threading import Lock

from nsanic.libs.mult_log import NLogger


class SingleTon(type):
    """ 单例元类 加线程锁"""
    __instances = {}

    def __new__(cls, *args, **kwargs):
        """ 当类实例化时调用该方法 """
        if not cls.__instances.get(cls):
            # 多线程环境下，若类实例构造方法中有IO任务，则需要加锁让获取锁的线程完成实例化，后续线程访问直接从__instances获取
            with Lock():
                cls.__instances[cls] = super().__new__(*args, **kwargs)
        return cls.__instances[cls]


def with_meta(meta, base_class=object):
    """
    meta: 元类, type或type的派生类, 用于创建类
    等价于继承 base_class, metaclass=meta
    example:
        ```
        class Test(with_meta(type, superclass)):
            pass
        ```
    """
    return meta("NewBase", (base_class,), {})


def singleton(cls):
    """单实例装饰器 加线程锁"""
    ins_map = {}

    def ins(*args, **kwargs):
        if cls not in ins_map:
            with Lock():
                ins_map[cls] = cls(*args, **kwargs)
        return ins_map[cls]
    return ins


def aio_catcher(log: NLogger = None):
    """协程异常捕获器"""

    def expt(fun):
        @wraps(fun)
        async def wrap_fun(*args, **kwargs):
            try:
                return await fun(*args, **kwargs)
            except Exception as err:
                err_str = f"执行出错：{err}\n{ traceback.format_exc()}"
                log.error(err_str) if log else print(err_str)
        return wrap_fun
    return expt


def catcher(log: NLogger = None):
    """常规异常捕获器"""

    def expt(fun):
        @wraps(fun)
        def wrap_fun(*args, **kwargs):
            try:
                return fun(*args, **kwargs)
            except Exception as err:
                err_str = f"执行出错：{err}\n{traceback.format_exc()}"
                log.error(err_str) if log else print(err_str)
        return wrap_fun
    return expt


def aio_runtime(log: NLogger = None):
    """协程 时间检查"""

    def out_runtime(fun):
        @wraps(fun)
        async def wrap_fun(*args, **kwargs):
            s_time = time.time()
            req = await fun(*args, **kwargs)
            e_time = time.time()
            info_str = f'程序执行时间: {(e_time - s_time) * 1000}毫秒'
            log.info(info_str) if log else print(info_str)
            return req
        return wrap_fun
    return out_runtime


def runtime(log: NLogger = None):
    """常规时间检查"""

    def out_runtime(fun):
        @wraps(fun)
        def wrap_fun(*args, **kwargs):
            s_time = time.time()
            req = fun(*args, **kwargs)
            e_time = time.time()
            info_str = f'程序执行时间: {(e_time - s_time) * 1000}毫秒'
            log.info(info_str) if log else print(info_str)
            return req
        return wrap_fun
    return out_runtime



import asyncio
from asyncio import queues
from typing import Union

from nsanic.libs import tool, tool_ws
from nsanic.libs.component import LogMeta
from nsanic.libs.rds_client import RdsClient


class WsConnector(LogMeta):
    """Websocket连接管理器"""
    OFFLINE_QUEUE = queues.Queue()
    _WS_MAP = {}

    @classmethod
    def all_ws(cls):
        return cls._WS_MAP.values()

    @classmethod
    def get_ws(cls, key: Union[str, int, bytes]):
        return cls._WS_MAP.get(key)

    @classmethod
    async def save_to_histories(cls, receiver, msg: [dict, list]):
        pass

    @classmethod
    async def send_histories(cls, receiver):
        pass

    @classmethod
    async def check_old_ws(cls, key):
        return None

    @classmethod
    async def set_ws(cls, ukey: Union[str, int, bytes], ws, key_info: str = None):
        """更新连接，如果存在旧连接会关闭旧连接"""
        old_ws = cls.get_ws(ukey)
        cls._WS_MAP[ukey] = ws
        old_ws and (await old_ws.close())

    @classmethod
    async def close_ws(cls, ukey: Union[str, int, bytes], msg: Union[str, bytes] = None):
        ws = cls.get_ws(ukey)
        if ws:
            cls._WS_MAP.pop(ukey)
            try:
                msg and (await ws.send(msg))
                await ws.close()
            except Exception as err:
                _ = err

    @classmethod
    async def close_all(cls, reason: str = None):
        for ws in cls._WS_MAP.values():
            try:
                reason and await ws.send(reason)
                await ws.close()
            except Exception as err:
                _ = err
        cls._WS_MAP.clear()

    @classmethod
    async def offline(cls, ukey: Union[str, int, bytes]):
        await cls.close_ws(ukey)


class WsRdsConnector(WsConnector):
    """基于Redis缓存控制的Websocket连接管理器"""

    _rds: RdsClient = None
    CHANNEL_LIST = set()
    MAIN_CHANNEL = ''
    fun_pack_msg = tool_ws.pack_msg
    WS_ONLINE_KEY = 'WS_ONLINE_INFO'
    HISTORY_KEY = 'HISTORIES_USER_MESSAGE'
    HISTORY_COUNT = 50
    HISTORY_TYPE = 1
    '''历史消息存储类型，1--hash 针对数量少存储少的小体量消息，容易清理，2--LIST队列，针对存储量大的或者数量大的大体量消息，清理耗时但性能更好'''
    HISTORY_SEND_WAY = 1
    '''历史消息发送方式，1--一次性发所有历史消息 针对数量少存储少的小体量消息，2--将历史消息逐条发，针对存储量大的或者数量大的大体量消息'''

    @classmethod
    def set_conf(cls, channel: str, proc_name = ''):
        suffix = f"_{proc_name}" if proc_name else ""
        cls.MAIN_CHANNEL = f'NOTICE_{channel}{suffix}'
        cls.CHANNEL_LIST.add(cls.MAIN_CHANNEL)

    @classmethod
    def init_loop(cls, **kwargs):
        cls._rds = kwargs.get('rds')
        loop = kwargs.get('loop')
        if not loop:
            loop = asyncio.get_running_loop()
        loop.create_task(cls.init_rds_listen())

    @classmethod
    async def init_rds_listen(cls):
        if not cls._rds:
            cls._rds = RdsClient.clt()
        try:
            sub_obj = await cls._rds.pub_sub(list(cls.CHANNEL_LIST))
        except Exception as err:
            cls.loginfo(f'订阅消息频道出错:{cls.CHANNEL_LIST},5秒后即将重启监听任务：{err}')
            await asyncio.sleep(5)
            return await cls.init_rds_listen()
        while 1:
            try:
                t, c, m = await sub_obj.parse_response()
            except Exception as err:
                cls.logerr(f'监听任务出错:{cls.CHANNEL_LIST},10秒后即将重启监听任务：{err}')
                await asyncio.sleep(10)
                break
            if isinstance(t, bytes):
                t, c, m = t.decode(), c.decode(), (m.decode() if isinstance(m, bytes) else m)
            if t == 'message':
                if c == cls.MAIN_CHANNEL:
                    arg = cls.parse_main_channel(m)
                    arg and (await cls.on_receive_main_channel(*arg))
                    continue
                arg = cls.parse_channel_msg(c, m)
                arg and (await cls.on_receive_channel(c, *arg))
        # 监听失败 5秒后重启监听任务
        await asyncio.sleep(5)
        await cls.init_rds_listen()

    @classmethod
    def parse_main_channel(cls, msg):
        """主通道监听消消息解析 如有需要可重写"""
        msg = tool.json_parse(msg)
        if (not msg) or not (isinstance(msg, dict)):
            return
        c_type, c_code, data, u = msg.get('t'), msg.get('c'), msg.get('d'), (msg.get('u') or -1)
        if not c_type:
            return
        return [c_type, c_code, data, u]

    @classmethod
    async def on_receive_main_channel(cls, cmd_type, cmd_code, data: dict, receiver):
        if not receiver:
            return
        if receiver == -1:
            for ws in cls.all_ws():
                try:
                    ws and (await ws.send(
                        cls.fun_pack_msg(cmd_type, cmd_code, data=data)))
                except Exception as err:
                    cls.loginfo(f'公共消息推送失败：{err}')
                    pass
            return
        ws = cls.get_ws(receiver)
        if not ws:
            return await cls.save_to_histories(receiver, cls.fun_pack_msg(cmd_type, cmd_code, data=data))
        try:
            return await ws.send(cls.fun_pack_msg(cmd_type, cmd_code, data=data))
        except Exception as err:
            cls.loginfo(f'发送目标消息失败,receiver:{receiver},data:{data},错误信息：{err}')
        return await cls.save_to_histories(receiver, cls.fun_pack_msg(cmd_type, cmd_code, data=data))

    @classmethod
    async def save_to_histories(cls, receiver, msg: [dict, list]):
        if not msg:
            return
        if not isinstance(msg, str):
            msg = tool.json_encode(msg, u_byte=True)
        if cls.HISTORY_TYPE == 1:
            if isinstance(msg, dict):
                msg = [msg]
            msg_list = await cls._rds.get_hash(cls.HISTORY_KEY, receiver, jsparse=True)
            msg_list = msg_list.extend(msg) if msg_list else msg
            out_count = len(msg_list) - cls.HISTORY_COUNT
            if out_count > 0:
                msg_list = msg_list[out_count:]
            return await cls._rds.set_hash(cls.HISTORY_KEY, receiver, msg_list)
        key_name = f"{cls.HISTORY_KEY}:{receiver}"
        await cls._rds.qlpush(key_name, [msg] if isinstance(msg, dict) else msg)
        out_count = ((await cls._rds.qlen(key_name)) or 0) - cls.HISTORY_COUNT
        if out_count > 0:
            _ = await cls._rds.qrpop(key_name, count=out_count)

    @classmethod
    async def __send_msgs(cls, ws, msgs: Union[list, dict]):
        if not ws:
            return
        if cls.HISTORY_SEND_WAY or isinstance(msgs, dict):
            try:
                await ws.send(tool.json_encode(msgs))
                return True
            except Exception as err:
                cls.loginfo(f"推送历史消息失败，将重新入队: {err}")
            return
        for index, msg in enumerate(msgs):
            try:
                await ws.send(tool.json_encode(msg))
                continue
            except Exception as err:
                cls.loginfo(f"推送历史消息失败，将重新入队: {err}")
            return index
        return True

    @classmethod
    async def check_old_ws(cls, key):
        return await cls._rds.get_hash(cls.WS_ONLINE_KEY, key)

    @classmethod
    async def send_histories(cls, receiver):
        if receiver not in cls._WS_MAP:
            return
        if cls.HISTORY_TYPE == 1:
            msg_list = await cls._rds.get_hash(cls.HISTORY_KEY, receiver, jsparse=True)
            if not msg_list:
                return
            ws = cls._WS_MAP.get(receiver)
            sta = cls.__send_msgs(ws, msg_list)
            sta and (await cls._rds.drop_hash(cls.HISTORY_KEY, receiver))
        key_name = f"{cls.HISTORY_KEY}:{receiver}"
        msg_list = await cls._rds.qrpop(key_name, count=cls.HISTORY_COUNT)
        if not msg_list:
            return
        msg_list = [tool.json_parse(i) for i in msg_list]
        ws = cls._WS_MAP.get(receiver)
        idx = await cls.__send_msgs(ws, msg_list)
        if isinstance(idx, int):
            await cls.save_to_histories(receiver, [tool.json_parse(i) for i in msg_list[idx:]])
        if idx is None:
            await cls.save_to_histories(receiver, msg_list)

    @classmethod
    async def set_ws(cls, ukey: Union[str, int, bytes], ws, key_info: str = None):
        """更新连接，如果存在旧连接会关闭旧连接"""
        old_ws = cls.get_ws(ukey)
        cls._WS_MAP[ukey] = ws
        if old_ws:
            cls._rds and await cls._rds.drop_hash(
                cls.WS_ONLINE_KEY, getattr(old_ws, 'ukey', ''))
            await old_ws.close()
        key_info and cls._rds and await cls._rds.set_hash(cls.WS_ONLINE_KEY, ukey, key_info)

    @classmethod
    async def close_ws(cls, ukey: Union[str, int, bytes], msg: Union[str, bytes] = None, del_key=True):
        ws = cls.get_ws(ukey)
        if ws:
            cls._WS_MAP.pop(ukey)
            msg and await ws.send(msg)
            await ws.close()
        del_key and cls._rds and await cls._rds.drop_hash(cls.WS_ONLINE_KEY, ukey)

    @classmethod
    async def close_all(cls, reason: str = None):
        for ws in cls._WS_MAP.values():
            try:
                reason and await ws.send(reason)
                await ws.close()
            except Exception as err:
                _ = err
        cls._WS_MAP.clear()
        cls._rds and await cls._rds.del_item(cls.WS_ONLINE_KEY)

    @classmethod
    def parse_channel_msg(cls, channel: str, msg: dict):
        """解析其它通道的消息请重写该方法"""
        raise Exception('未实现其它频道的消息解析方法')

    @classmethod
    async def on_receive_channel(cls, channel: str, cmd_type, cmd_code, data: dict, receiver):
        """
        处理其它频道的消息
        :param channel: 频道标识
        :param cmd_type: 消息类型
        :param cmd_code: 消息标号
        :param data: 消息数据
        :param receiver: 接收者
        """
        raise Exception('未实现其它频道的消息处理方法')


class HeaderSet:

    RESP_TYPE = {
        'JSON': 'application/json',
        'MSGPACK': 'text/msgpack',
        'TEXT': 'text/plain',
        'HTML': 'text/html',
        'PROTO': 'application/octet-stream',
    }

    @classmethod
    def out(cls, conf):
        tp = cls.RESP_TYPE.get(conf.RESP_TYPE)
        if not tp:
            raise Exception('无效的响应类型配置')
        if conf.RESP_TYPE != 'MSGPACK':
            tp += f';charset={conf.CHARSET}'
        headers = {
            'Content-Type': tp,
            'Access-Control-Allow-Methods': ','.join(conf.ALLOW_METHOD),
            'Access-Control-Max-Age': conf.ORIGIN_MAX_AGE,
            'Access-Control-Allow-Headers': '*' if '*' in conf.ALLOW_HEADER else ','.join(conf.ALLOW_HEADER),
            'Access-Control-Allow-Origin': '*' if '*' in conf.ALLOW_ORIGIN else ','.join(conf.ALLOW_ORIGIN),
            'Access-Control-Allow-Credentials': conf.ALLOW_CREDENTIALS
        }
        return headers
